@@ -5,7 +5,8 @@ rae@cs.brynmawr.edu
 -}
 
 {-# LANGUAGE CPP, MultiParamTypeClasses, FunctionalDependencies,
-             TypeSynonymInstances, FlexibleInstances, LambdaCase #-}
+             TypeSynonymInstances, FlexibleInstances, LambdaCase,
+             ScopedTypeVariables #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -23,8 +24,10 @@ rae@cs.brynmawr.edu
 
 module Language.Haskell.TH.Desugar (
   -- * Desugared data types
-  DExp(..), DLetDec(..), DPat(..), DType(..), DKind, DCxt, DPred,
-  DTyVarBndr(..), DMatch(..), DClause(..), DDec(..),
+  DExp(..), DLetDec(..), DPat(..),
+  DType(..), DForallTelescope(..), DKind, DCxt, DPred,
+  DTyVarBndr(..), DTyVarBndrSpec, DTyVarBndrUnit, Specificity(..),
+  DMatch(..), DClause(..), DDec(..),
   DDerivClause(..), DDerivStrategy(..), DPatSynDir(..), DPatSynType,
   Overlap(..), PatSynArgs(..), NewOrData(..),
   DTypeFamilyHead(..), DFamilyResultSig(..), InjectivityAnn(..),
@@ -40,7 +43,7 @@ module Language.Haskell.TH.Desugar (
   -- * Main desugaring functions
   dsExp, dsDecs, dsType, dsInfo,
   dsPatOverExp, dsPatsOverExp, dsPatX,
-  dsLetDecs, dsTvb, dsCxt,
+  dsLetDecs, dsTvb, dsTvbSpec, dsTvbUnit, dsCxt,
   dsCon, dsForeign, dsPragma, dsRuleBndr,
 
   -- ** Secondary desugaring functions
@@ -65,10 +68,11 @@ module Language.Haskell.TH.Desugar (
   -- * Reification
   reifyWithWarning,
 
-  -- | The following definitions allow you to register a list of
-  -- @Dec@s to be used in reification queries.
-  withLocalDeclarations, dsReify,
+  -- ** Local reification
+  -- $localReification
+  withLocalDeclarations, dsReify, dsReifyType,
   reifyWithLocals_maybe, reifyWithLocals, reifyFixityWithLocals,
+  reifyTypeWithLocals_maybe, reifyTypeWithLocals,
   lookupValueNameWithLocals, lookupTypeNameWithLocals,
   mkDataNameWithLocals, mkTypeNameWithLocals,
   reifyNameSpace,
@@ -97,8 +101,15 @@ module Language.Haskell.TH.Desugar (
 #if __GLASGOW_HASKELL__ >= 800
   bindIP,
 #endif
-  unravel, conExistentialTvbs, mkExtraDKindBinders,
-  dTyVarBndrToDType, toposortTyVarsOf,
+  mkExtraDKindBinders, dTyVarBndrToDType, changeDTVFlags, toposortTyVarsOf,
+
+  -- ** 'FunArgs' and 'VisFunArg'
+  FunArgs(..), ForallTelescope(..), VisFunArg(..),
+  filterVisFunArgs, ravelType, unravelType,
+
+  -- ** 'DFunArgs' and 'DVisFunArg'
+  DFunArgs(..), DVisFunArg(..),
+  filterDVisFunArgs, ravelDType, unravelDType,
 
   -- ** 'TypeArg'
   TypeArg(..), applyType, filterTANormals, unfoldType,
@@ -110,12 +121,12 @@ module Language.Haskell.TH.Desugar (
   extractBoundNamesStmt, extractBoundNamesDec, extractBoundNamesPat
   ) where
 
+import Language.Haskell.TH.Datatype.TyVarBndr
 import Language.Haskell.TH.Desugar.AST
 import Language.Haskell.TH.Desugar.Core
 import Language.Haskell.TH.Desugar.Expand
 import Language.Haskell.TH.Desugar.FV
 import Language.Haskell.TH.Desugar.Match
-import qualified Language.Haskell.TH.Desugar.OSet as OS
 import Language.Haskell.TH.Desugar.Reify
 import Language.Haskell.TH.Desugar.Subst
 import Language.Haskell.TH.Desugar.Sweeten
@@ -130,10 +141,22 @@ import qualified Data.Map as M
 import qualified Data.Set as S
 import Prelude hiding ( exp )
 
+#if __GLASGOW_HASKELL__ < 710
+import Control.Applicative
+#endif
+
 -- | This class relates a TH type with its th-desugar type and allows
 -- conversions back and forth. The functional dependency goes only one
--- way because `Type` and `Kind` are type synonyms, but they desugar
--- to different types.
+-- way because we define the following instances on old versions of GHC:
+--
+-- @
+-- instance 'Desugar' 'TyVarBndrSpec' 'DTyVarBndrSpec'
+-- instance 'Desugar' 'TyVarBndrUnit' 'DTyVarBndrUnit'
+-- @
+--
+-- Prior to GHC 9.0, 'TyVarBndrSpec' and 'TyVarBndrUnit' are simply type
+-- synonyms for 'TyVarBndr', so making the functional dependencies
+-- bidirectional would cause these instances to be rejected.
 class Desugar th ds | ds -> th where
   desugar :: DsMonad q => th -> q ds
   sweeten :: ds -> th
@@ -150,9 +173,36 @@ instance Desugar Cxt DCxt where
   desugar = dsCxt
   sweeten = cxtToTH
 
-instance Desugar TyVarBndr DTyVarBndr where
+#if __GLASGOW_HASKELL__ >= 900
+-- | This instance is only @flag@-polymorphic on GHC 9.0 or later, since
+-- previous versions of GHC do not equip 'TyVarBndr' with a @flag@ type
+-- parameter. As a result, we define two separate instances for 'DTyVarBndr'
+-- on older GHCs:
+--
+-- @
+-- instance 'Desugar' 'TyVarBndrSpec' 'DTyVarBndrSpec'
+-- instance 'Desugar' 'TyVarBndrUnit' 'DTyVarBndrUnit'
+-- @
+instance Desugar (TyVarBndr flag) (DTyVarBndr flag) where
   desugar = dsTvb
   sweeten = tvbToTH
+#else
+-- | This instance monomorphizes the @flag@ parameter of 'DTyVarBndr' since
+-- pre-9.0 versions of GHC do not equip 'TyVarBndr' with a @flag@ type
+-- parameter. There is also a corresponding instance for
+-- 'TyVarBndrUnit'/'DTyVarBndrUnit'.
+instance Desugar TyVarBndrSpec DTyVarBndrSpec where
+  desugar = dsTvbSpec
+  sweeten = tvbToTH
+
+-- | This instance monomorphizes the @flag@ parameter of 'DTyVarBndr' since
+-- pre-9.0 versions of GHC do not equip 'TyVarBndr' with a @flag@ type
+-- parameter. There is also a corresponding instance for
+-- 'TyVarBndrSpec'/'DTyVarBndrSpec'.
+instance Desugar TyVarBndrUnit DTyVarBndrUnit where
+  desugar = dsTvbUnit
+  sweeten = tvbToTH
+#endif
 
 instance Desugar [Dec] [DDec] where
   desugar = dsDecs
@@ -219,38 +269,38 @@ flattenDValD other_dec = return [other_dec]
 --
 -- instead of returning one binding for @X1@ and another binding for @X2@.
 --
--- 'getRecordSelectors' attempts to filter out \"naughty\" record selectors
--- whose types mention existentially quantified type variables. But see the
--- documentation for 'conExistentialTvbs' for limitations to this approach.
-
--- See https://github.com/goldfirere/singletons/issues/180 for an example where
--- the latter behavior can bite you.
-
-getRecordSelectors :: DsMonad q
-                   => DType        -- ^ the type of the argument
-                   -> [DCon]
-                   -> q [DLetDec]
-getRecordSelectors arg_ty cons = merge_let_decs `fmap` concatMapM get_record_sels cons
+-- 'getRecordSelectors' does not attempt to filter out \"naughty\" record
+-- selectors—that is, records whose field types mention existentially
+-- quantified type variables that do not appear in the constructor's return
+-- type. Here is an example of a naughty record selector:
+--
+-- @
+-- data Some :: (Type -> Type) -> Type where
+--   MkSome :: { getSome :: f a } -> Some f
+-- @
+--
+-- GHC itself will not allow the use of @getSome@ as a top-level function due
+-- to its type @f a@ mentioning the existential variable @a@, but
+-- 'getRecordSelectors' will return it nonetheless. Ultimately, this design
+-- choice is a practical one, as detecting which type variables are existential
+-- in Template Haskell is difficult in the general case.
+getRecordSelectors :: DsMonad q => [DCon] -> q [DLetDec]
+getRecordSelectors cons = merge_let_decs `fmap` concatMapM get_record_sels cons
   where
-    get_record_sels con@(DCon con_tvbs _ con_name con_fields con_ret_ty) =
+    get_record_sels (DCon con_tvbs _ con_name con_fields con_ret_ty) =
       case con_fields of
         DRecC fields -> go fields
         DNormalC{}   -> return []
         where
           go fields = do
             varName <- qNewName "field"
-            con_ex_tvbs <- conExistentialTvbs arg_ty con
-            let con_univ_tvbs  = deleteFirstsBy ((==) `on` dtvbName) con_tvbs con_ex_tvbs
-                con_ex_tvb_set = OS.fromList $ map dtvbName con_ex_tvbs
-                forall'        = DForallT con_univ_tvbs []
-                num_pats       = length fields
             return $ concat
-              [ [ DSigD name (forall' $ DArrowT `DAppT` con_ret_ty `DAppT` field_ty)
-                , DFunD name [DClause [DConP con_name (mk_field_pats n num_pats varName)]
+              [ [ DSigD name $ DForallT (DForallInvis con_tvbs)
+                             $ DArrowT `DAppT` con_ret_ty `DAppT` field_ty
+                , DFunD name [DClause [DConP con_name
+                                         (mk_field_pats n (length fields) varName)]
                                       (DVarE varName)] ]
               | ((name, _strict, field_ty), n) <- zip fields [0..]
-              , OS.null (fvDType field_ty `OS.intersection` con_ex_tvb_set)
-                  -- exclude "naughty" selectors
               ]
 
     mk_field_pats :: Int -> Int -> Name -> [DPat]
@@ -332,76 +382,70 @@ getRecordSelectors arg_ty cons = merge_let_decs `fmap` concatMapM get_record_sel
 -- are fresh type variable names.
 --
 -- This expands kind synonyms if necessary.
-mkExtraDKindBinders :: DsMonad q => DKind -> q [DTyVarBndr]
-mkExtraDKindBinders = expandType >=> mkExtraDKindBinders'
+mkExtraDKindBinders :: forall q. DsMonad q => DKind -> q [DTyVarBndrUnit]
+mkExtraDKindBinders k = do
+  k' <- expandType k
+  let (fun_args, _) = unravelDType k'
+      vis_fun_args  = filterDVisFunArgs fun_args
+  mapM mk_tvb vis_fun_args
+  where
+    mk_tvb :: DVisFunArg -> q DTyVarBndrUnit
+    mk_tvb (DVisFADep tvb) = return tvb
+    mk_tvb (DVisFAAnon ki) = DKindedTV <$> qNewName "a" <*> return () <*> return ki
 
--- | Returns all of a constructor's existentially quantified type variable
--- binders.
---
--- Detecting the presence of existentially quantified type variables in the
--- context of Template Haskell is quite involved. Here is an example that
--- we will use to explain how this works:
---
--- @
--- data family Foo a b
--- data instance Foo (Maybe a) b where
---   MkFoo :: forall x y z. x -> y -> z -> Foo (Maybe x) [z]
--- @
---
--- In @MkFoo@, @x@ is universally quantified, whereas @y@ and @z@ are
--- existentially quantified. Note that @MkFoo@ desugars (in Core) to
--- something like this:
---
--- @
--- data instance Foo (Maybe a) b where
---   MkFoo :: forall a b y z. (b ~ [z]). a -> y -> z -> Foo (Maybe a) b
--- @
---
--- Here, we can see that @a@ appears in the desugared return type (it is a
--- simple alpha-renaming of @x@), so it is universally quantified. On the other
--- hand, neither @y@ nor @z@ appear in the desugared return type, so they are
--- existentially quantified.
---
--- This analysis would not have been possible without knowing what the original
--- data declaration's type was (in this case, @Foo (Maybe a) b@), which is why
--- we require it as an argument. Our algorithm for detecting existentially
--- quantified variables is not too different from what was described above:
--- we match the constructor's return type with the original data type, forming
--- a substitution, and check which quantified variables are not part of the
--- domain of the substitution.
---
--- Be warned: this may overestimate which variables are existentially
--- quantified when kind variables are involved. For instance, consider this
--- example:
---
--- @
--- data S k (a :: k)
--- data T a where
---   MkT :: forall k (a :: k). { foo :: Proxy (a :: k), bar :: S k a } -> T a
--- @
---
--- Here, the kind variable @k@ does not appear syntactically in the return type
--- @T a@, so 'conExistentialTvbs' would mistakenly flag @k@ as existential.
---
--- There are various tricks we could employ to improve this, but ultimately,
--- making this behave correctly with respect to @PolyKinds@ 100% of the time
--- would amount to performing kind inference in Template Haskell, which is
--- quite difficult. For the sake of simplicity, we have decided to stick with
--- a dumb-but-predictable syntactic check.
-conExistentialTvbs :: DsMonad q
-                   => DType -- ^ The type of the original data declaration
-                   -> DCon
-                   -> q [DTyVarBndr]
-conExistentialTvbs data_ty (DCon tvbs _ _ _ ret_ty) = do
-  data_ty' <- expandType data_ty
-  ret_ty'  <- expandType ret_ty
-  case matchTy YesIgnore ret_ty' data_ty' of
-    Nothing -> fail $ showString "Unable to match type "
-                    . showsPrec 11 ret_ty'
-                    . showString " with "
-                    . showsPrec 11 data_ty'
-                    $ ""
-    Just gadtSubt -> return [ tvb
-                            | tvb <- tvbs
-                            , M.notMember (dtvbName tvb) gadtSubt
-                            ]
+{- $localReification
+
+@template-haskell@ reification functions like 'reify' and 'qReify', as well as
+@th-desugar@'s 'reifyWithWarning', only look through declarations that either
+(1) have already been typechecked in the current module, or (2) are in scope
+because of imports. We refer to this as /global/ reification. Sometimes,
+however, you may wish to reify declarations that have been quoted but not
+yet been typechecked, such as in the following example:
+
+@
+example :: IO ()
+example = putStrLn
+  $(do decs <- [d| data Foo = MkFoo |]
+       info <- 'reify' (mkName \"Foo\")
+       stringE $ pprint info)
+@
+
+Because @Foo@ only exists in a TH quote, it is not available globally. As a
+result, the call to @'reify' (mkName \"Foo\")@ will fail.
+
+To make this sort of example possible, @th-desugar@ extends global reification
+with /local/ reification. A function that performs local reification (such
+as 'dsReify', 'reifyWithLocals', or similar functions that have a 'DsMonad'
+context) looks through both typechecked (or imported) declarations /and/ quoted
+declarations that are currently in scope. One can add quoted declarations in
+the current scope by using the 'withLocalDeclarations' function. Here is an
+example of how to repair the example above using 'withLocalDeclarations':
+
+@
+example2 :: IO ()
+example2 = putStrLn
+  $(do decs <- [d| data Foo = MkFoo |]
+       info <- 'withLocalDeclarations' decs $
+                 'reifyWithLocals' (mkName \"Foo\")
+       stringE $ pprint info)
+@
+
+Note that 'withLocalDeclarations' should only be used to add quoted
+declarations with names that are not duplicates of existing global or local
+declarations. Adding duplicate declarations through 'withLocalDeclarations'
+is undefined behavior and should be avoided. This is unlikely to happen if
+you are only using 'withLocalDeclarations' in conjunction with TH quotes,
+however. For instance, this is /not/ an example of duplicate declarations:
+
+@
+data T = MkT1
+
+$(do decs <- [d| data T = MkT2 |]
+     info <- 'withLocalDeclarations' decs ...
+     ...)
+@
+
+The quoted @data T = MkT2@ does not conflict with the top-level @data T = Mk1@
+since declaring a data type within TH quotes gives it a fresh, unique name that
+distinguishes it from any other data types already in scope.
+-}
