@@ -15,8 +15,8 @@ module Language.Haskell.TH.Desugar.Core where
 import Prelude hiding (mapM, foldl, foldr, all, elem, exp, concatMap, and)
 
 import Language.Haskell.TH hiding (match, clause, cxt)
+import Language.Haskell.TH.Datatype.TyVarBndr
 import Language.Haskell.TH.Syntax hiding (lift)
-import Language.Haskell.TH.Datatype ( resolveTypeSynonyms )
 
 #if __GLASGOW_HASKELL__ < 709
 import Control.Applicative
@@ -78,7 +78,10 @@ dsExp (InfixE (Just lhs) op (Just rhs)) =
 dsExp (UInfixE _ _ _) =
   fail "Cannot desugar unresolved infix operators."
 dsExp (ParensE exp) = dsExp exp
-dsExp (LamE pats exp) = dsLam pats =<< dsExp exp
+dsExp (LamE pats exp) = do
+  exp' <- dsExp exp
+  (pats', exp'') <- dsPatsOverExp pats exp'
+  mkDLamEFromDPats pats' exp''
 dsExp (LamCaseE matches) = do
   x <- newUniqueName "x"
   matches' <- dsMatches x matches
@@ -106,7 +109,11 @@ dsExp (CaseE exp matches) = do
   matches' <- dsMatches scrutinee matches
   return $ DLetE [DValD (DVarP scrutinee) exp'] $
            DCaseE (DVarE scrutinee) matches'
-dsExp (DoE stmts) = dsDoStmts stmts
+#if __GLASGOW_HASKELL__ >= 900
+dsExp (DoE mb_mod stmts) = dsDoStmts mb_mod stmts
+#else
+dsExp (DoE        stmts) = dsDoStmts Nothing stmts
+#endif
 dsExp (CompE stmts) = dsComp stmts
 dsExp (ArithSeqE (FromR exp)) = DAppE (DVarE 'enumFrom) <$> dsExp exp
 dsExp (ArithSeqE (FromThenR exp1 exp2)) =
@@ -265,7 +272,7 @@ ds_tup tuple_data_name mb_exps = do
   if null section_vars
      then return tup_body -- If this isn't a tuple section,
                           -- don't create a lambda.
-     else dsLam (map VarP section_vars) tup_body
+     else mkDLamEFromDPats (map DVarP section_vars) tup_body
   where
     -- If dealing with an empty field in a tuple section (Nothing), create a
     -- unique name and return Left. These names will be used to construct the
@@ -286,40 +293,22 @@ ds_tup tuple_data_name mb_exps = do
     apply_tup_body f (Left n)  = f `DAppE` DVarE n
     apply_tup_body f (Right e) = f `DAppE` e
 
--- | Desugar a lambda expression, where the body has already been desugared
-dsLam :: DsMonad q => [Pat] -> DExp -> q DExp
-dsLam = mkLam stripVarP_maybe dsPatsOverExp
-
 -- | Convert a list of 'DPat' arguments and a 'DExp' body into a 'DLamE'. This
 -- is needed since 'DLamE' takes a list of 'Name's for its bound variables
 -- instead of 'DPat's, so some reorganization is needed.
-mkDLamEFromDPats :: DsMonad q => [DPat] -> DExp -> q DExp
-mkDLamEFromDPats = mkLam stripDVarP_maybe (\pats exp -> return (pats, exp))
-  where
-    stripDVarP_maybe :: DPat -> Maybe Name
-    stripDVarP_maybe (DVarP n) = Just n
-    stripDVarP_maybe _          = Nothing
-
--- | Generalizes 'dsLam' and 'mkDLamEFromDPats' to work over an arbitrary
--- pattern type.
-mkLam :: DsMonad q
-      => (pat -> Maybe Name) -- ^ Should return @'Just' n@ if the argument is a
-                             --   variable pattern, and 'Nothing' otherwise.
-      -> ([pat] -> DExp -> q ([DPat], DExp))
-                             -- ^ Should process a list of @pat@ arguments and
-                             --   a 'DExp' body. (This might do some internal
-                             --   reorganization if there are as-patterns, as
-                             --   in the case of 'dsPatsOverExp'.)
-      -> [pat] -> DExp -> q DExp
-mkLam mb_strip_var_pat process_pats_over_exp pats exp
-  | Just names <- mapM mb_strip_var_pat pats
+mkDLamEFromDPats :: Quasi q => [DPat] -> DExp -> q DExp
+mkDLamEFromDPats pats exp
+  | Just names <- mapM stripDVarP_maybe pats
   = return $ DLamE names exp
   | otherwise
   = do arg_names <- replicateM (length pats) (newUniqueName "arg")
        let scrutinee = mkTupleDExp (map DVarE arg_names)
-       (pats', exp') <- process_pats_over_exp pats exp
-       let match = DMatch (mkTupleDPat pats') exp'
+           match     = DMatch (mkTupleDPat pats) exp
        return $ DLamE arg_names (DCaseE scrutinee [match])
+  where
+    stripDVarP_maybe :: DPat -> Maybe Name
+    stripDVarP_maybe (DVarP n) = Just n
+    stripDVarP_maybe _          = Nothing
 
 -- | Desugar a list of matches for a @case@ statement
 dsMatches :: DsMonad q
@@ -415,23 +404,27 @@ dsGuardStmts (RecS {} : _) _ _ = fail "th-desugar currently does not support Rec
 #endif
 
 -- | Desugar the @Stmt@s in a @do@ expression
-dsDoStmts :: DsMonad q => [Stmt] -> q DExp
-dsDoStmts [] = impossible "do-expression ended with something other than bare statement."
-dsDoStmts [NoBindS exp] = dsExp exp
-dsDoStmts (BindS pat exp : rest) = do
-  rest' <- dsDoStmts rest
-  dsBindS exp pat rest' "do expression"
-dsDoStmts (LetS decs : rest) = do
-  (decs', ip_binder) <- dsLetDecs decs
-  rest' <- dsDoStmts rest
-  return $ DLetE decs' $ ip_binder rest'
-dsDoStmts (NoBindS exp : rest) = do
-  exp' <- dsExp exp
-  rest' <- dsDoStmts rest
-  return $ DAppE (DAppE (DVarE '(>>)) exp') rest'
-dsDoStmts (ParS _ : _) = impossible "Parallel comprehension in a do-statement."
+dsDoStmts :: forall q. DsMonad q => Maybe ModName -> [Stmt] -> q DExp
+dsDoStmts mb_mod = go
+  where
+    go :: [Stmt] -> q DExp
+    go [] = impossible "do-expression ended with something other than bare statement."
+    go [NoBindS exp] = dsExp exp
+    go (BindS pat exp : rest) = do
+      rest' <- go rest
+      dsBindS mb_mod exp pat rest' "do expression"
+    go (LetS decs : rest) = do
+      (decs', ip_binder) <- dsLetDecs decs
+      rest' <- go rest
+      return $ DLetE decs' $ ip_binder rest'
+    go (NoBindS exp : rest) = do
+      exp' <- dsExp exp
+      rest' <- go rest
+      let sequence_name = mk_qual_do_name mb_mod '(>>)
+      return $ DAppE (DAppE (DVarE sequence_name) exp') rest'
+    go (ParS _ : _) = impossible "Parallel comprehension in a do-statement."
 #if __GLASGOW_HASKELL__ >= 807
-dsDoStmts (RecS {} : _) = fail "th-desugar currently does not support RecursiveDo"
+    go (RecS {} : _) = fail "th-desugar currently does not support RecursiveDo"
 #endif
 
 -- | Desugar the @Stmt@s in a list or monad comprehension
@@ -440,7 +433,7 @@ dsComp [] = impossible "List/monad comprehension ended with something other than
 dsComp [NoBindS exp] = DAppE (DVarE 'return) <$> dsExp exp
 dsComp (BindS pat exp : rest) = do
   rest' <- dsComp rest
-  dsBindS exp pat rest' "monad comprehension"
+  dsBindS Nothing exp pat rest' "monad comprehension"
 dsComp (LetS decs : rest) = do
   (decs', ip_binder) <- dsLetDecs decs
   rest' <- dsComp rest
@@ -452,7 +445,7 @@ dsComp (NoBindS exp : rest) = do
 dsComp (ParS stmtss : rest) = do
   (pat, exp) <- dsParComp stmtss
   rest' <- dsComp rest
-  DAppE (DAppE (DVarE '(>>=)) exp) <$> dsLam [pat] rest'
+  DAppE (DAppE (DVarE '(>>=)) exp) <$> mkDLamEFromDPats [pat] rest'
 #if __GLASGOW_HASKELL__ >= 807
 dsComp (RecS {} : _) = fail "th-desugar currently does not support RecursiveDo"
 #endif
@@ -464,12 +457,13 @@ dsComp (RecS {} : _) = fail "th-desugar currently does not support RecursiveDo"
 -- 'MonadFail' or 'Monad', depending on whether the @MonadFailDesugaring@
 -- language extension is enabled or not. (On GHCs older than 8.0, 'fail' from
 -- 'Monad' is always used.)
-dsBindS :: forall q. DsMonad q => Exp -> Pat -> DExp -> String -> q DExp
-dsBindS bind_arg_exp success_pat success_exp ctxt = do
+dsBindS :: forall q. DsMonad q
+        => Maybe ModName -> Exp -> Pat -> DExp -> String -> q DExp
+dsBindS mb_mod bind_arg_exp success_pat success_exp ctxt = do
   bind_arg_exp' <- dsExp bind_arg_exp
   (success_pat', success_exp') <- dsPatOverExp success_pat success_exp
   is_univ_pat <- isUniversalPattern success_pat'
-  let bind_into = DAppE (DAppE (DVarE '(>>=)) bind_arg_exp')
+  let bind_into = DAppE (DAppE (DVarE bind_name) bind_arg_exp')
   if is_univ_pat
      then bind_into <$> mkDLamEFromDPats [success_pat'] success_exp'
      else do arg_name  <- newUniqueName "arg"
@@ -481,36 +475,46 @@ dsBindS bind_arg_exp success_pat success_exp ctxt = do
                    DLitE (StringL $ "Pattern match failure in " ++ ctxt)
                ]
   where
+    bind_name = mk_qual_do_name mb_mod '(>>=)
+
     mk_fail_name :: q Name
 #if __GLASGOW_HASKELL__ >= 807
     -- GHC 8.8 deprecates the MonadFailDesugaring extension since its effects
     -- are always enabled. Furthermore, MonadFailDesugaring is no longer
     -- enabled by default, so simply use MonadFail.fail. (That happens to
     -- be the same as Prelude.fail in 8.8+.)
-    mk_fail_name = return 'MonadFail.fail
+    mk_fail_name = return fail_MonadFail_name
 #elif __GLASGOW_HASKELL__ >= 800
     mk_fail_name = do
       mfd <- qIsExtEnabled MonadFailDesugaring
-      return $ if mfd then 'MonadFail.fail else 'Prelude.fail
+      return $ if mfd then fail_MonadFail_name else fail_Prelude_name
 #else
-    mk_fail_name = return 'Prelude.fail
+    mk_fail_name = return fail_Prelude_name
+#endif
+
+#if __GLASGOW_HASKELL__ >= 800
+    fail_MonadFail_name = mk_qual_do_name mb_mod 'MonadFail.fail
+#endif
+
+#if __GLASGOW_HASKELL__ < 807
+    fail_Prelude_name = mk_qual_do_name mb_mod 'Prelude.fail
 #endif
 
 -- | Desugar the contents of a parallel comprehension.
 --   Returns a @Pat@ containing a tuple of all bound variables and an expression
 --   to produce the values for those variables
-dsParComp :: DsMonad q => [[Stmt]] -> q (Pat, DExp)
+dsParComp :: DsMonad q => [[Stmt]] -> q (DPat, DExp)
 dsParComp [] = impossible "Empty list of parallel comprehension statements."
 dsParComp [r] = do
   let rv = foldMap extractBoundNamesStmt r
   dsR <- dsComp (r ++ [mk_tuple_stmt rv])
-  return (mk_tuple_pat rv, dsR)
+  return (mk_tuple_dpat rv, dsR)
 dsParComp (q : rest) = do
   let qv = foldMap extractBoundNamesStmt q
   (rest_pat, rest_exp) <- dsParComp rest
   dsQ <- dsComp (q ++ [mk_tuple_stmt qv])
   let zipped = DAppE (DAppE (DVarE 'mzip) dsQ) rest_exp
-  return (ConP (tupleDataName 2) [mk_tuple_pat qv, rest_pat], zipped)
+  return (DConP (tupleDataName 2) [mk_tuple_dpat qv, rest_pat], zipped)
 
 -- helper function for dsParComp
 mk_tuple_stmt :: OSet Name -> Stmt
@@ -518,23 +522,23 @@ mk_tuple_stmt name_set =
   NoBindS (mkTupleExp (F.foldr ((:) . VarE) [] name_set))
 
 -- helper function for dsParComp
-mk_tuple_pat :: OSet Name -> Pat
-mk_tuple_pat name_set =
-  mkTuplePat (F.foldr ((:) . VarP) [] name_set)
+mk_tuple_dpat :: OSet Name -> DPat
+mk_tuple_dpat name_set =
+  mkTupleDPat (F.foldr ((:) . DVarP) [] name_set)
 
 -- | Desugar a pattern, along with processing a (desugared) expression that
 -- is the entire scope of the variables bound in the pattern.
 dsPatOverExp :: DsMonad q => Pat -> DExp -> q (DPat, DExp)
 dsPatOverExp pat exp = do
   (pat', vars) <- runWriterT $ dsPat pat
-  let name_decs = uncurry (zipWith (DValD . DVarP)) $ unzip vars
+  let name_decs = map (uncurry (DValD . DVarP)) vars
   return (pat', maybeDLetE name_decs exp)
 
 -- | Desugar multiple patterns. Like 'dsPatOverExp'.
 dsPatsOverExp :: DsMonad q => [Pat] -> DExp -> q ([DPat], DExp)
 dsPatsOverExp pats exp = do
   (pats', vars) <- runWriterT $ mapM dsPat pats
-  let name_decs = uncurry (zipWith (DValD . DVarP)) $ unzip vars
+  let name_decs = map (uncurry (DValD . DVarP)) vars
   return (pats', maybeDLetE name_decs exp)
 
 -- | Desugar a pattern, returning a list of (Name, DExp) pairs of extra
@@ -692,9 +696,9 @@ dsDec (NewtypeD cxt n tvbs con derivings) =
   dsDataDec Newtype cxt n tvbs Nothing [con] derivings
 #endif
 dsDec (TySynD n tvbs ty) =
-  (:[]) <$> (DTySynD n <$> mapM dsTvb tvbs <*> dsType ty)
+  (:[]) <$> (DTySynD n <$> mapM dsTvbUnit tvbs <*> dsType ty)
 dsDec (ClassD cxt n tvbs fds decs) =
-  (:[]) <$> (DClassD <$> dsCxt cxt <*> pure n <*> mapM dsTvb tvbs
+  (:[]) <$> (DClassD <$> dsCxt cxt <*> pure n <*> mapM dsTvbUnit tvbs
                      <*> pure fds <*> dsDecs decs)
 #if __GLASGOW_HASKELL__ >= 711
 dsDec (InstanceD over cxt ty decs) =
@@ -711,12 +715,12 @@ dsDec d@(PragmaD {}) = dsTopLevelLetDec d
 dsDec (OpenTypeFamilyD tfHead) =
   (:[]) <$> (DOpenTypeFamilyD <$> dsTypeFamilyHead tfHead)
 dsDec (DataFamilyD n tvbs m_k) =
-  (:[]) <$> (DDataFamilyD n <$> mapM dsTvb tvbs <*> mapM dsType m_k)
+  (:[]) <$> (DDataFamilyD n <$> mapM dsTvbUnit tvbs <*> mapM dsType m_k)
 #else
 dsDec (FamilyD TypeFam n tvbs m_k) = do
   (:[]) <$> (DOpenTypeFamilyD <$> dsTypeFamilyHead n tvbs m_k)
 dsDec (FamilyD DataFam n tvbs m_k) =
-  (:[]) <$> (DDataFamilyD n <$> mapM dsTvb tvbs <*> mapM dsType m_k)
+  (:[]) <$> (DDataFamilyD n <$> mapM dsTvbUnit tvbs <*> mapM dsType m_k)
 #endif
 #if __GLASGOW_HASKELL__ >= 807
 dsDec (DataInstD cxt mtvbs lhs mk cons derivings) =
@@ -774,13 +778,16 @@ dsDec (DefaultSigD n ty) = (:[]) <$> (DDefaultSigD n <$> dsType ty)
 #if __GLASGOW_HASKELL__ >= 807
 dsDec (ImplicitParamBindD {}) = impossible "Non-`let`-bound implicit param binding"
 #endif
+#if __GLASGOW_HASKELL__ >= 809
+dsDec (KiSigD n ki) = (:[]) <$> (DKiSigD n <$> dsType ki)
+#endif
 
 -- | Desugar a 'DataD' or 'NewtypeD'.
 dsDataDec :: DsMonad q
-          => NewOrData -> Cxt -> Name -> [TyVarBndr]
+          => NewOrData -> Cxt -> Name -> [TyVarBndrUnit]
           -> Maybe Kind -> [Con] -> [DerivingClause] -> q [DDec]
 dsDataDec nd cxt n tvbs mk cons derivings = do
-  tvbs' <- mapM dsTvb tvbs
+  tvbs' <- mapM dsTvbUnit tvbs
   let h98_tvbs = case mk of
                    -- If there's an explicit return kind, we're dealing with a
                    -- GADT, so this argument goes unused in dsCon.
@@ -794,10 +801,10 @@ dsDataDec nd cxt n tvbs mk cons derivings = do
 
 -- | Desugar a 'DataInstD' or a 'NewtypeInstD'.
 dsDataInstDec :: DsMonad q
-              => NewOrData -> Cxt -> Name -> Maybe [TyVarBndr] -> [TypeArg]
+              => NewOrData -> Cxt -> Name -> Maybe [TyVarBndrUnit] -> [TypeArg]
               -> Maybe Kind -> [Con] -> [DerivingClause] -> q [DDec]
 dsDataInstDec nd cxt n mtvbs tys mk cons derivings = do
-  mtvbs' <- mapM (mapM dsTvb) mtvbs
+  mtvbs' <- mapM (mapM dsTvbUnit) mtvbs
   tys'   <- mapM dsTypeArg tys
   let lhs' = applyDType (DConT n) tys'
       h98_tvbs =
@@ -817,28 +824,17 @@ dsDataInstDec nd cxt n mtvbs tys mk cons derivings = do
                            <*> concatMapM (dsCon h98_tvbs h98_fam_inst_type) cons
                            <*> mapM dsDerivClause derivings)
 
--- Like mkExtraDKindBinders, but accepts a Maybe Kind
--- argument instead of DKind.
-mkExtraKindBinders :: DsMonad q => Maybe Kind -> q [DTyVarBndr]
-mkExtraKindBinders =
-  maybe (pure (DConT typeKindName)) (runQ . resolveTypeSynonyms >=> dsType)
-    >=> mkExtraDKindBinders'
-
--- | Like mkExtraDKindBinders, but assumes kind synonyms have been expanded.
-mkExtraDKindBinders' :: Quasi q => DKind -> q [DTyVarBndr]
-mkExtraDKindBinders' = mkExtraKindBindersGeneric unravel DKindedTV
-
 #if __GLASGOW_HASKELL__ > 710
 -- | Desugar a @FamilyResultSig@
 dsFamilyResultSig :: DsMonad q => FamilyResultSig -> q DFamilyResultSig
 dsFamilyResultSig NoSig          = return DNoSig
 dsFamilyResultSig (KindSig k)    = DKindSig <$> dsType k
-dsFamilyResultSig (TyVarSig tvb) = DTyVarSig <$> dsTvb tvb
+dsFamilyResultSig (TyVarSig tvb) = DTyVarSig <$> dsTvbUnit tvb
 
 -- | Desugar a @TypeFamilyHead@
 dsTypeFamilyHead :: DsMonad q => TypeFamilyHead -> q DTypeFamilyHead
 dsTypeFamilyHead (TypeFamilyHead n tvbs result inj)
-  = DTypeFamilyHead n <$> mapM dsTvb tvbs
+  = DTypeFamilyHead n <$> mapM dsTvbUnit tvbs
                       <*> dsFamilyResultSig result
                       <*> pure inj
 
@@ -847,12 +843,12 @@ typeFamilyHeadName (TypeFamilyHead n _ _ _) = n
 #else
 -- | Desugar bits and pieces into a 'DTypeFamilyHead'
 dsTypeFamilyHead :: DsMonad q
-                 => Name -> [TyVarBndr] -> Maybe Kind -> q DTypeFamilyHead
+                 => Name -> [TyVarBndrUnit] -> Maybe Kind -> q DTypeFamilyHead
 dsTypeFamilyHead n tvbs m_kind = do
   result_sig <- case m_kind of
     Nothing -> return DNoSig
     Just k  -> DKindSig <$> dsType k
-  DTypeFamilyHead n <$> mapM dsTvb tvbs
+  DTypeFamilyHead n <$> mapM dsTvbUnit tvbs
                     <*> pure result_sig
                     <*> pure Nothing
 #endif
@@ -969,10 +965,10 @@ dsTopLevelLetDec = fmap (map DLetDec . fst) . dsLetDec
 -- we require passing these as arguments. (If we desugar an actual GADT
 -- constructor, these arguments are ignored.)
 dsCon :: DsMonad q
-      => [DTyVarBndr] -- ^ The universally quantified type variables
-                      --   (used if desugaring a non-GADT constructor).
-      -> DType        -- ^ The original data declaration's type
-                      --   (used if desugaring a non-GADT constructor).
+      => [DTyVarBndrUnit] -- ^ The universally quantified type variables
+                          --   (used if desugaring a non-GADT constructor).
+      -> DType            -- ^ The original data declaration's type
+                          --   (used if desugaring a non-GADT constructor).
       -> Con -> q [DCon]
 dsCon univ_dtvbs data_type con = do
   dcons' <- dsCon' con
@@ -980,8 +976,10 @@ dsCon univ_dtvbs data_type con = do
     case m_gadt_type of
       Nothing ->
         let ex_dtvbs   = dtvbs
-            expl_dtvbs = univ_dtvbs ++ ex_dtvbs
-            impl_dtvbs = toposortTyVarsOf $ mapMaybe extractTvbKind expl_dtvbs in
+            expl_dtvbs = changeDTVFlags SpecifiedSpec univ_dtvbs ++
+                         ex_dtvbs
+            impl_dtvbs = changeDTVFlags SpecifiedSpec $
+                         toposortTyVarsOf $ mapMaybe extractTvbKind expl_dtvbs in
         DCon (impl_dtvbs ++ expl_dtvbs) dcxt n fields data_type
       Just gadt_type ->
         let univ_ex_dtvbs = dtvbs in
@@ -996,7 +994,7 @@ dsCon univ_dtvbs data_type con = do
 -- * If returning Nothing, we're dealing with a non-GADT constructor, so
 --   the returned DTyVarBndrs are the existentials only.
 dsCon' :: DsMonad q
-       => Con -> q [(Name, [DTyVarBndr], DCxt, DConFields, Maybe DType)]
+       => Con -> q [(Name, [DTyVarBndrSpec], DCxt, DConFields, Maybe DType)]
 dsCon' (NormalC n stys) = do
   dtys <- mapM dsBangType stys
   return [(n, [], [], DNormalC False dtys, Nothing)]
@@ -1008,7 +1006,7 @@ dsCon' (InfixC sty1 n sty2) = do
   dty2 <- dsBangType sty2
   return [(n, [], [], DNormalC True [dty1, dty2], Nothing)]
 dsCon' (ForallC tvbs cxt con) = do
-  dtvbs <- mapM dsTvb tvbs
+  dtvbs <- mapM dsTvbSpec tvbs
   dcxt <- dsCxt cxt
   dcons' <- dsCon' con
   return $ flip map dcons' $ \(n, dtvbs', dcxt', fields, m_gadt_type) ->
@@ -1066,7 +1064,7 @@ dsPragma (SpecialiseP n ty m_inl phases) = DSpecialiseP n <$> dsType ty
 dsPragma (SpecialiseInstP ty)            = DSpecialiseInstP <$> dsType ty
 #if __GLASGOW_HASKELL__ >= 807
 dsPragma (RuleP str mtvbs rbs lhs rhs phases)
-                                         = DRuleP str <$> mapM (mapM dsTvb) mtvbs
+                                         = DRuleP str <$> mapM (mapM dsTvbUnit) mtvbs
                                                       <*> mapM dsRuleBndr rbs
                                                       <*> dsExp lhs
                                                       <*> dsExp rhs
@@ -1098,7 +1096,7 @@ dsRuleBndr (TypedRuleVar n ty) = DTypedRuleVar n <$> dsType ty
 -- this information prior to GHC 8.8.
 dsTySynEqn :: DsMonad q => Name -> TySynEqn -> q DTySynEqn
 dsTySynEqn _ (TySynEqn mtvbs lhs rhs) =
-  DTySynEqn <$> mapM (mapM dsTvb) mtvbs <*> dsType lhs <*> dsType rhs
+  DTySynEqn <$> mapM (mapM dsTvbUnit) mtvbs <*> dsType lhs <*> dsType rhs
 #else
 -- | Desugar a @TySynEqn@. (Available only with GHC 7.8+)
 dsTySynEqn :: DsMonad q => Name -> TySynEqn -> q DTySynEqn
@@ -1143,7 +1141,14 @@ dsClauses n clauses@(Clause outer_pats _ _ : _) = do
 
 -- | Desugar a type
 dsType :: DsMonad q => Type -> q DType
-dsType (ForallT tvbs preds ty) = DForallT <$> mapM dsTvb tvbs <*> dsCxt preds <*> dsType ty
+#if __GLASGOW_HASKELL__ >= 900
+-- See Note [Gracefully handling linear types]
+dsType (MulArrowT `AppT` _) = return DArrowT
+dsType MulArrowT = fail "Cannot desugar exotic uses of linear types."
+#endif
+dsType (ForallT tvbs preds ty) =
+  mkDForallConstrainedT <$> (DForallInvis <$> mapM dsTvbSpec tvbs)
+                        <*> dsCxt preds <*> dsType ty
 dsType (AppT t1 t2) = DAppT <$> dsType t1 <*> dsType t2
 dsType (SigT ty ki) = DSigT <$> dsType ty <*> dsType ki
 dsType (VarT name) = return $ DVarT name
@@ -1181,11 +1186,72 @@ dsType (ImplicitParamT n t) = do
   t' <- dsType t
   return $ DConT ''IP `DAppT` DLitT (StrTyLit n) `DAppT` t'
 #endif
+#if __GLASGOW_HASKELL__ >= 809
+dsType (ForallVisT tvbs ty) =
+  DForallT <$> (DForallVis <$> mapM dsTvbUnit tvbs) <*> dsType ty
+#endif
 
--- | Desugar a @TyVarBndr@
-dsTvb :: DsMonad q => TyVarBndr -> q DTyVarBndr
-dsTvb (PlainTV n) = return $ DPlainTV n
-dsTvb (KindedTV n k) = DKindedTV n <$> dsType k
+#if __GLASGOW_HASKELL__ >= 900
+-- | Desugar a 'TyVarBndr'.
+dsTvb :: DsMonad q => TyVarBndr_ flag -> q (DTyVarBndr flag)
+dsTvb (PlainTV n flag)    = return $ DPlainTV n flag
+dsTvb (KindedTV n flag k) = DKindedTV n flag <$> dsType k
+#else
+-- | Desugar a 'TyVarBndr' with a particular @flag@.
+dsTvb :: DsMonad q => flag -> TyVarBndr -> q (DTyVarBndr flag)
+dsTvb flag (PlainTV n)    = return $ DPlainTV n flag
+dsTvb flag (KindedTV n k) = DKindedTV n flag <$> dsType k
+#endif
+
+{-
+Note [Gracefully handling linear types]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Per the README, th-desugar does not currently support linear types.
+Unfortunately, we cannot simply reject all occurrences of
+multiplicity-polymorphic function arrows (i.e., MulArrowT), as it is possible
+for "non-linear" code to contain them when reified. For example, the type of a
+Haskell98 data constructor such as `Just` will be reified as
+
+  a #-> Maybe a
+
+In terms of the TH AST, that is:
+
+  MulArrowT `AppT` PromotedConT 'One `AppT` VarT a `AppT` (ConT ''Maybe `AppT` VarT a)
+
+Therefore, in order to desugar these sorts of types, we have to do *something*
+with MulArrowT. The approach that th-desugar takes is to pretend that all
+multiplicity-polymorphic function arrows are actually ordinary function arrows
+(->) when desugaring types. In other words, whenever th-desugar sees
+(MulArrowT `AppT` m), for any particular value of `m`, it will turn it into
+DArrowT.
+
+This approach is enough to gracefully handle most uses of MulArrowT, as TH
+reification always generates MulArrowT applied to some particular multiplicity
+(as of GHC 9.0, at least). It's conceivable that some wily user could manually
+construct a TH AST containing MulArrowT in a different position, but since this
+situation is rare, we simply throw an error in such cases.
+
+We adopt a similar stance in L.H.TH.Desugar.Reify when locally reifying the
+types of data constructors: since th-desugar doesn't currently support linear
+types, we pretend as if MulArrowT does not exist. As a result, the type of
+`Just` would be locally reified as `a -> Maybe a`, not `a #-> Maybe a`.
+-}
+
+-- | Desugar a 'TyVarBndrSpec'.
+dsTvbSpec :: DsMonad q => TyVarBndrSpec -> q DTyVarBndrSpec
+#if __GLASGOW_HASKELL__ >= 900
+dsTvbSpec = dsTvb
+#else
+dsTvbSpec = dsTvb SpecifiedSpec
+#endif
+
+-- | Desugar a 'TyVarBndrUnit'.
+dsTvbUnit :: DsMonad q => TyVarBndrUnit -> q DTyVarBndrUnit
+#if __GLASGOW_HASKELL__ >= 900
+dsTvbUnit = dsTvb
+#else
+dsTvbUnit = dsTvb ()
+#endif
 
 -- | Desugar a @Cxt@
 dsCxt :: DsMonad q => Cxt -> q DCxt
@@ -1245,12 +1311,7 @@ dsPred (EqualP t1 t2) = do
 dsPred t
   | Just ts <- splitTuple_maybe t
   = concatMapM dsPred ts
-dsPred (ForallT tvbs cxt p) = do
-  ps' <- dsPred p
-  case ps' of
-    [p'] -> (:[]) <$> (DForallT <$> mapM dsTvb tvbs <*> dsCxt cxt <*> pure p')
-    _    -> fail "Cannot desugar constraint tuples in the body of a quantified constraint"
-              -- See Trac #15334.
+dsPred (ForallT tvbs cxt p) = dsForallPred tvbs cxt p
 dsPred (AppT t1 t2) = do
   [p1] <- dsPred t1   -- tuples can't be applied!
   (:[]) <$> DAppT p1 <$> dsType t2
@@ -1298,6 +1359,23 @@ dsPred (ImplicitParamT n t) = do
   t' <- dsType t
   return [DConT ''IP `DAppT` DLitT (StrTyLit n) `DAppT` t']
 #endif
+#if __GLASGOW_HASKELL__ >= 809
+dsPred t@(ForallVisT {}) =
+  impossible $ "Visible dependent quantifier seen as head of constraint: " ++ show t
+#endif
+#if __GLASGOW_HASKELL__ >= 900
+dsPred MulArrowT = impossible "Linear arrow seen as head of constraint."
+#endif
+
+-- | Desugar a quantified constraint.
+dsForallPred :: DsMonad q => [TyVarBndrSpec] -> Cxt -> Pred -> q DCxt
+dsForallPred tvbs cxt p = do
+  ps' <- dsPred p
+  case ps' of
+    [p'] -> (:[]) <$> (mkDForallConstrainedT <$>
+                         (DForallInvis <$> mapM dsTvbSpec tvbs) <*> dsCxt cxt <*> pure p')
+    _    -> fail "Cannot desugar constraint tuples in the body of a quantified constraint"
+              -- See GHC #15334.
 #endif
 
 -- | Like 'reify', but safer and desugared. Uses local declarations where
@@ -1305,6 +1383,19 @@ dsPred (ImplicitParamT n t) = do
 dsReify :: DsMonad q => Name -> q (Maybe DInfo)
 dsReify = traverse dsInfo <=< reifyWithLocals_maybe
 
+-- | Like 'reifyType', but safer and desugared. Uses local declarations where
+-- available.
+dsReifyType :: DsMonad q => Name -> q (Maybe DType)
+dsReifyType = traverse dsType <=< reifyTypeWithLocals_maybe
+
+-- Given a list of `forall`ed type variable binders and a context, construct
+-- a DType using DForallT and DConstrainedT as appropriate. The phrase
+-- "as appropriate" is used because DConstrainedT will not be used if the
+-- context is empty, per Note [Desugaring and sweetening ForallT].
+mkDForallConstrainedT :: DForallTelescope -> DCxt -> DType -> DType
+mkDForallConstrainedT tele ctxt ty =
+  DForallT tele $ if null ctxt then ty else DConstrainedT ctxt ty
+
 -- create a list of expressions in the same order as the fields in the first argument
 -- but with the values as given in the second argument
 -- if a field is missing from the second argument, use the corresponding expression
@@ -1355,11 +1446,6 @@ mkTupleDPat :: [DPat] -> DPat
 mkTupleDPat [pat] = pat
 mkTupleDPat pats = DConP (tupleDataName (length pats)) pats
 
--- | Make a tuple 'Pat' from a list of 'Pat's. Avoids using a 1-tuple.
-mkTuplePat :: [Pat] -> Pat
-mkTuplePat [pat] = pat
-mkTuplePat pats = ConP (tupleDataName (length pats)) pats
-
 -- | Is this pattern guaranteed to match?
 isUniversalPattern :: DsMonad q => DPat -> q Bool
 isUniversalPattern (DLitP {}) = return False
@@ -1411,9 +1497,9 @@ filterDTANormals = mapMaybe getDTANormal
     getDTANormal (DTyArg {})   = Nothing
 
 -- | Convert a 'DTyVarBndr' into a 'DType'
-dTyVarBndrToDType :: DTyVarBndr -> DType
-dTyVarBndrToDType (DPlainTV a)    = DVarT a
-dTyVarBndrToDType (DKindedTV a k) = DVarT a `DSigT` k
+dTyVarBndrToDType :: DTyVarBndr flag -> DType
+dTyVarBndrToDType (DPlainTV a _)    = DVarT a
+dTyVarBndrToDType (DKindedTV a _ k) = DVarT a `DSigT` k
 
 -- | Extract the underlying 'DType' or 'DKind' from a 'DTypeArg'. This forgets
 -- information about whether a type is a normal argument or not, so use with
@@ -1437,7 +1523,7 @@ strictToBang = id
 
 -- Take a data type name (which does not belong to a data family) and
 -- apply it to its type variable binders to form a DType.
-nonFamilyDataReturnType :: Name -> [DTyVarBndr] -> DType
+nonFamilyDataReturnType :: Name -> [DTyVarBndrUnit] -> DType
 nonFamilyDataReturnType con_name =
   applyDType (DConT con_name) . map (DTANormal . dTyVarBndrToDType)
 
@@ -1453,7 +1539,7 @@ dataFamInstReturnType fam_name = applyDType (DConT fam_name)
 -- arguments. We accomplish this by taking the free variables of the types
 -- and performing a reverse topological sort on them to ensure that the
 -- returned list is well scoped.
-dataFamInstTvbs :: [DTypeArg] -> [DTyVarBndr]
+dataFamInstTvbs :: [DTypeArg] -> [DTyVarBndrUnit]
 dataFamInstTvbs = toposortTyVarsOf . map probablyWrongUnDTypeArg
 
 -- | Take a list of 'DType's, find their free variables, and sort them in
@@ -1469,7 +1555,7 @@ dataFamInstTvbs = toposortTyVarsOf . map probablyWrongUnDTypeArg
 --
 -- On older GHCs, this takes measures to avoid returning explicitly bound
 -- kind variables, which was not possible before @TypeInType@.
-toposortTyVarsOf :: [DType] -> [DTyVarBndr]
+toposortTyVarsOf :: [DType] -> [DTyVarBndrUnit]
 toposortTyVarsOf tys =
   let freeVars :: [Name]
       freeVars = F.toList $ foldMap fvDType tys
@@ -1478,8 +1564,8 @@ toposortTyVarsOf tys =
       varKindSigs = foldMap go_ty tys
         where
           go_ty :: DType -> Map Name DKind
-          go_ty (DForallT tvbs ctxt t) =
-            go_tvbs tvbs (foldMap go_ty ctxt `mappend` go_ty t)
+          go_ty (DForallT tele t) = go_tele tele (go_ty t)
+          go_ty (DConstrainedT ctxt t) = foldMap go_ty ctxt `mappend` go_ty t
           go_ty (DAppT t1 t2) = go_ty t1 `mappend` go_ty t2
           go_ty (DAppKindT t k) = go_ty t `mappend` go_ty k
           go_ty (DSigT t k) =
@@ -1493,12 +1579,16 @@ toposortTyVarsOf tys =
           go_ty (DLitT {}) = mempty
           go_ty DWildCardT = mempty
 
-          go_tvbs :: [DTyVarBndr] -> Map Name DKind -> Map Name DKind
+          go_tele :: DForallTelescope -> Map Name DKind -> Map Name DKind
+          go_tele (DForallVis   tvbs) = go_tvbs tvbs
+          go_tele (DForallInvis tvbs) = go_tvbs tvbs
+
+          go_tvbs :: [DTyVarBndr flag] -> Map Name DKind -> Map Name DKind
           go_tvbs tvbs m = foldr go_tvb m tvbs
 
-          go_tvb :: DTyVarBndr -> Map Name DKind -> Map Name DKind
-          go_tvb (DPlainTV n)    m = M.delete n m
-          go_tvb (DKindedTV n k) m = M.delete n m `mappend` go_ty k
+          go_tvb :: DTyVarBndr flag -> Map Name DKind -> Map Name DKind
+          go_tvb (DPlainTV n _)    m = M.delete n m
+          go_tvb (DKindedTV n _ k) m = M.delete n m `mappend` go_ty k
 
       -- | Do a topological sort on a list of tyvars,
       --   so that binders occur before occurrences
@@ -1552,7 +1642,7 @@ toposortTyVarsOf tys =
         maybe S.empty (OS.toSet . fvDType)
                       (M.lookup n varKindSigs)
       ascribeWithKind n =
-        maybe (DPlainTV n) (DKindedTV n) (M.lookup n varKindSigs)
+        maybe (DPlainTV n ()) (DKindedTV n ()) (M.lookup n varKindSigs)
 
       -- An annoying wrinkle: GHCs before 8.0 don't support explicitly
       -- quantifying kinds, so something like @forall k (a :: k)@ would be
@@ -1571,20 +1661,81 @@ toposortTyVarsOf tys =
      filter (not . isKindBinderOnOldGHCs) $
      scopedSort freeVars
 
-dtvbName :: DTyVarBndr -> Name
-dtvbName (DPlainTV n)    = n
-dtvbName (DKindedTV n _) = n
+dtvbName :: DTyVarBndr flag -> Name
+dtvbName (DPlainTV n _)    = n
+dtvbName (DKindedTV n _ _) = n
 
--- | Decompose a function type into its type variables, its context, its
--- argument types, and its result type.
-unravel :: DType -> ([DTyVarBndr], [DPred], [DType], DType)
-unravel (DForallT tvbs cxt ty) =
-  let (tvbs', cxt', tys, res) = unravel ty in
-  (tvbs ++ tvbs', cxt ++ cxt', tys, res)
-unravel (DAppT (DAppT DArrowT t1) t2) =
-  let (tvbs, cxt, tys, res) = unravel t2 in
-  (tvbs, cxt, t1 : tys, res)
-unravel t = ([], [], [], t)
+-- @mk_qual_do_name mb_mod orig_name@ will simply return @orig_name@ if
+-- @mb_mod@ is Nothing. If @mb_mod@ is @Just mod_@, then a new 'Name' will be
+-- returned that uses @mod_@ as the new module prefix. This is useful for
+-- emulating the behavior of the @QualifiedDo@ extension, which adds module
+-- prefixes to functions such as ('>>=') and ('>>').
+mk_qual_do_name :: Maybe ModName -> Name -> Name
+mk_qual_do_name mb_mod orig_name = case mb_mod of
+  Nothing   -> orig_name
+  Just mod_ -> Name (OccName (nameBase orig_name)) (NameQ mod_)
+
+-- | Reconstruct an arrow 'DType' from its argument and result types.
+ravelDType :: DFunArgs -> DType -> DType
+ravelDType DFANil                 res = res
+ravelDType (DFAForalls tele args) res = DForallT tele (ravelDType args res)
+ravelDType (DFACxt cxt args)      res = DConstrainedT cxt (ravelDType args res)
+ravelDType (DFAAnon t args)       res = DAppT (DAppT DArrowT t) (ravelDType args res)
+
+-- | Decompose a function 'DType' into its arguments (the 'DFunArgs') and its
+-- result type (the 'DType).
+unravelDType :: DType -> (DFunArgs, DType)
+unravelDType (DForallT tele ty) =
+  let (args, res) = unravelDType ty in
+  (DFAForalls tele args, res)
+unravelDType (DConstrainedT cxt ty) =
+  let (args, res) = unravelDType ty in
+  (DFACxt cxt args, res)
+unravelDType (DAppT (DAppT DArrowT t1) t2) =
+  let (args, res) = unravelDType t2 in
+  (DFAAnon t1 args, res)
+unravelDType t = (DFANil, t)
+
+-- | The list of arguments in a function 'DType'.
+data DFunArgs
+  = DFANil
+    -- ^ No more arguments.
+  | DFAForalls DForallTelescope DFunArgs
+    -- ^ A series of @forall@ed type variables followed by a dot (if
+    --   'ForallInvis') or an arrow (if 'ForallVis'). For example,
+    --   the type variables @a1 ... an@ in @forall a1 ... an. r@.
+  | DFACxt DCxt DFunArgs
+    -- ^ A series of constraint arguments followed by @=>@. For example,
+    --   the @(c1, ..., cn)@ in @(c1, ..., cn) => r@.
+  | DFAAnon DType DFunArgs
+    -- ^ An anonymous argument followed by an arrow. For example, the @a@
+    --   in @a -> r@.
+  deriving (Eq, Show, Typeable, Data, Generic)
+
+-- | A /visible/ function argument type (i.e., one that must be supplied
+-- explicitly in the source code). This is in contrast to /invisible/
+-- arguments (e.g., the @c@ in @c => r@), which are instantiated without
+-- the need for explicit user input.
+data DVisFunArg
+  = DVisFADep DTyVarBndrUnit
+    -- ^ A visible @forall@ (e.g., @forall a -> a@).
+  | DVisFAAnon DType
+    -- ^ An anonymous argument followed by an arrow (e.g., @a -> r@).
+  deriving (Eq, Show, Typeable, Data, Generic)
+
+-- | Filter the visible function arguments from a list of 'DFunArgs'.
+filterDVisFunArgs :: DFunArgs -> [DVisFunArg]
+filterDVisFunArgs DFANil = []
+filterDVisFunArgs (DFAForalls tele args) =
+  case tele of
+    DForallVis tvbs -> map DVisFADep tvbs ++ args'
+    DForallInvis _  -> args'
+  where
+    args' = filterDVisFunArgs args
+filterDVisFunArgs (DFACxt _ args) =
+  filterDVisFunArgs args
+filterDVisFunArgs (DFAAnon t args) =
+  DVisFAAnon t:filterDVisFunArgs args
 
 -- | Decompose an applied type into its individual components. For example, this:
 --
@@ -1601,16 +1752,29 @@ unfoldDType :: DType -> (DType, [DTypeArg])
 unfoldDType = go []
   where
     go :: [DTypeArg] -> DType -> (DType, [DTypeArg])
-    go acc (DForallT _ _ ty) = go acc ty
+    go acc (DForallT _ ty)   = go acc ty
     go acc (DAppT ty1 ty2)   = go (DTANormal ty2:acc) ty1
     go acc (DAppKindT ty ki) = go (DTyArg ki:acc) ty
     go acc (DSigT ty _)      = go acc ty
     go acc ty                = (ty, acc)
 
--- | Extract the kind from a 'TyVarBndr', if one is present.
-extractTvbKind :: DTyVarBndr -> Maybe DKind
-extractTvbKind (DPlainTV _) = Nothing
-extractTvbKind (DKindedTV _ k) = Just k
+-- | Extract the kind from a 'DTyVarBndr', if one is present.
+extractTvbKind :: DTyVarBndr flag -> Maybe DKind
+extractTvbKind (DPlainTV _ _)    = Nothing
+extractTvbKind (DKindedTV _ _ k) = Just k
+
+-- | Set the flag in a list of 'DTyVarBndr's. This is often useful in contexts
+-- where one needs to re-use a list of 'DTyVarBndr's from one flag setting to
+-- another flag setting. For example, in order to re-use the 'DTyVarBndr's bound
+-- by a 'DDataD' in a 'DForallT', one can do the following:
+--
+-- @
+-- case x of
+--   'DDataD' _ _ _ tvbs _ _ _ ->
+--     'DForallT' ('DForallInvis' ('changeDTVFlags' 'SpecifiedSpec' tvbs)) ...
+-- @
+changeDTVFlags :: newFlag -> [DTyVarBndr oldFlag] -> [DTyVarBndr newFlag]
+changeDTVFlags new_flag = map (new_flag <$)
 
 -- | Some functions in this module only use certain arguments on particular
 -- versions of GHC. Other versions of GHC (that don't make use of those
@@ -1620,3 +1784,51 @@ extractTvbKind (DKindedTV _ k) = Just k
 -- version of 'undefined'.)
 unusedArgument :: a
 unusedArgument = error "Unused"
+
+{-
+Note [Desugaring and sweetening ForallT]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+The ForallT constructor from template-haskell is tremendously awkward. Because
+ForallT contains both a list of type variable binders and constraint arguments,
+ForallT expressions can be ambiguous when one of these lists is empty. For
+example, consider this expression with no constraints:
+
+  ForallT [PlainTV a] [] (VarT a)
+
+What should this desugar to in th-desugar, which must maintain a clear
+separation between type variable binders and constraints? There are two
+possibilities:
+
+1. DForallT DForallInvis [DPlainTV a] (DVarT a)
+   (i.e., forall a. a)
+2. DForallT DForallInvis [DPlainTV a] (DConstrainedT [] (DVarT a))
+   (i.e., forall a. () => a)
+
+Template Haskell generally drops these empty lists when splicing Template
+Haskell expressions, so we would like to do the same in th-desugar to mimic
+TH's behavior as closely as possible. However, there are some situations where
+dropping empty lists of `forall`ed type variable binders can change the
+semantics of a program. For instance, contrast `foo :: forall. a -> a` (which
+is an error) with `foo :: a -> a` (which is fine). Therefore, we try to
+preserve empty `forall`s to the best of our ability.
+
+Here is an informal specification of how th-desugar should handle different sorts
+of ambiguity. First, a specification for desugaring.
+Let `tvbs` and `ctxt` be non-empty:
+
+* `ForallT tvbs [] ty` should desugar to `DForallT DForallInvis tvbs ty`.
+* `ForallT [] ctxt ty` should desguar to `DForallT DForallInvis [] (DConstrainedT ctxt ty)`.
+* `ForallT [] [] ty`   should desugar to `DForallT DForallInvis [] ty`.
+* For all other cases, just straightforwardly desugar
+  `ForallT tvbs ctxt ty` to `DForallT DForallInvis tvbs (DConstraintedT ctxt ty)`.
+
+For sweetening:
+
+* `DForallT DForallInvis tvbs (DConstrainedT ctxt ty)` should sweeten to `ForallT tvbs ctxt ty`.
+* `DForallT DForallInvis []   (DConstrainedT ctxt ty)` should sweeten to `ForallT [] ctxt ty`.
+* `DForallT DForallInvis tvbs (DConstrainedT [] ty)`   should sweeten to `ForallT tvbs [] ty`.
+* `DForallT DForallInvis []   (DConstrainedT [] ty)`   should sweeten to `ForallT [] [] ty`.
+* For all other cases, just straightforwardly sweeten
+  `DForallT DForallInvis tvbs ty` to `ForallT tvbs [] ty` and
+  `DConstrainedT ctxt ty` to `ForallT [] ctxt ty`.
+-}
