@@ -7,7 +7,7 @@ Allows for reification from a list of declarations, without looking a name
 up in the environment.
 -}
 
-{-# LANGUAGE CPP, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE CPP, GeneralizedNewtypeDeriving, ScopedTypeVariables #-}
 
 module Language.Haskell.TH.Desugar.Reify (
   -- * Reification
@@ -16,6 +16,10 @@ module Language.Haskell.TH.Desugar.Reify (
   -- ** Fixity reification
   qReifyFixity, reifyFixity, reifyFixityWithLocals, reifyFixityInDecs,
 
+  -- ** Type reification
+  qReifyType, reifyType,
+  reifyTypeWithLocals_maybe, reifyTypeWithLocals, reifyTypeInDecs,
+
   -- * Datatype lookup
   getDataD, dataConNameToCon, dataConNameToDataName,
 
@@ -28,6 +32,7 @@ module Language.Haskell.TH.Desugar.Reify (
   DsMonad(..), DsM, withLocalDeclarations
   ) where
 
+import Control.Applicative
 import qualified Control.Monad.Fail as Fail
 import Control.Monad.Reader
 import Control.Monad.State
@@ -35,14 +40,19 @@ import Control.Monad.Writer
 import Control.Monad.RWS
 import Control.Monad.Trans.Instances ()
 import qualified Data.Foldable as F
+#if __GLASGOW_HASKELL__ < 710
+import Data.Foldable (foldMap)
+#endif
 import Data.Function (on)
 import Data.List
+import qualified Data.Map as Map
+import Data.Map (Map)
 import Data.Maybe
-#if __GLASGOW_HASKELL__ < 709
-import Control.Applicative
-#endif
+import qualified Data.Set as Set
+import Data.Set (Set)
 
 import Language.Haskell.TH.Datatype
+import Language.Haskell.TH.Datatype.TyVarBndr
 import Language.Haskell.TH.Instances ()
 import Language.Haskell.TH.Syntax hiding ( lift )
 
@@ -90,7 +100,7 @@ reifyFail name =
 getDataD :: DsMonad q
          => String       -- ^ Print this out on failure
          -> Name         -- ^ Name of the datatype (@data@ or @newtype@) of interest
-         -> q ([TyVarBndr], [Con])
+         -> q ([TyVarBndrUnit], [Con])
 getDataD err name = do
   info <- reifyWithLocals name
   dec <- case info of
@@ -107,8 +117,8 @@ getDataD err name = do
     _ -> badDeclaration
   where
     go tvbs mk cons = do
-      k <- maybe (pure (ConT typeKindName)) (runQ . resolveTypeSynonyms) mk
-      extra_tvbs <- mkExtraKindBindersGeneric unravelType KindedTV k
+      let k = fromMaybe (ConT typeKindName) mk
+      extra_tvbs <- mkExtraKindBinders k
       let all_tvbs = tvbs ++ extra_tvbs
       return (all_tvbs, cons)
 
@@ -116,6 +126,31 @@ getDataD err name = do
           fail $ "The name (" ++ (show name) ++ ") refers to something " ++
                  "other than a datatype. " ++ err
 
+-- | Create new kind variable binder names corresponding to the return kind of
+-- a data type. This is useful when you have a data type like:
+--
+-- @
+-- data Foo :: forall k. k -> Type -> Type where ...
+-- @
+--
+-- But you want to be able to refer to the type @Foo a b@.
+-- 'mkExtraKindBinders' will take the kind @forall k. k -> Type -> Type@,
+-- discover that is has two visible argument kinds, and return as a result
+-- two new kind variable binders @[a :: k, b :: Type]@, where @a@ and @b@
+-- are fresh type variable names.
+--
+-- This expands kind synonyms if necessary.
+mkExtraKindBinders :: forall q. Quasi q => Kind -> q [TyVarBndrUnit]
+mkExtraKindBinders k = do
+  k' <- runQ $ resolveTypeSynonyms k
+  let (fun_args, _) = unravelType k'
+      vis_fun_args  = filterVisFunArgs fun_args
+  mapM mk_tvb vis_fun_args
+  where
+    mk_tvb :: VisFunArg -> q TyVarBndrUnit
+    mk_tvb (VisFADep tvb) = return tvb
+    mk_tvb (VisFAAnon ki) = kindedTV <$> qNewName "a" <*> return ki
+
 -- | From the name of a data constructor, retrive the datatype definition it
 -- is a part of.
 dataConNameToDataName :: DsMonad q => Name -> q Name
@@ -210,8 +245,10 @@ reifyInDecs n decs = snd `fmap` firstMatch (reifyInDec n decs) decs
 reifyFixityInDecs :: Name -> [Dec] -> Maybe Fixity
 reifyFixityInDecs n = firstMatch match_fixity
   where
-    match_fixity (InfixD fixity n') | n `nameMatches` n' = Just fixity
-    match_fixity _                                       = Nothing
+    match_fixity (InfixD fixity n')        | n `nameMatches` n'
+                                           = Just fixity
+    match_fixity (ClassD _ _ _ _ sub_decs) = firstMatch match_fixity sub_decs
+    match_fixity _                         = Nothing
 
 -- | A reified thing along with the name of that thing.
 type Named a = (Name, a)
@@ -279,7 +316,10 @@ reifyInDec n decs (ClassD _ ty_name tvbs _ sub_decs)
                                 reifyFixityInDecs n $ sub_decs ++ decs))
 #endif
 reifyInDec n decs (ClassD _ _ _ _ sub_decs)
-  | Just info <- firstMatch (reifyInDec n (sub_decs ++ decs)) sub_decs
+  | Just info <- firstMatch (reifyInDec n decs) sub_decs
+                 -- Important: don't pass (sub_decs ++ decs) to reifyInDec
+                 -- above, or else type family defaults can be confused for
+                 -- actual instances. See #134.
   = Just info
 #if __GLASGOW_HASKELL__ >= 711
 reifyInDec n decs (InstanceD _ _ _ sub_decs)
@@ -320,51 +360,85 @@ reifyInDec n decs (NewtypeInstD _ ty_name tys con _)
 reifyInDec _ _ _ = Nothing
 
 maybeReifyCon :: Name -> [Dec] -> Name -> [TypeArg] -> [Con] -> Maybe (Named Info)
-#if __GLASGOW_HASKELL__ > 710
 maybeReifyCon n _decs ty_name ty_args cons
   | Just (n', con) <- findCon n cons
-  = Just (n', DataConI n (maybeForallT tvbs [] $ con_to_type con) ty_name)
-#else
-maybeReifyCon n decs ty_name ty_args cons
-  | Just (n', con) <- findCon n cons
-  = Just (n', DataConI n (maybeForallT tvbs [] $ con_to_type con)
-                         ty_name fixity)
+    -- See Note [Use unSigType in maybeReifyCon]
+  , let full_con_ty = unSigType $ con_to_type h98_tvbs h98_res_ty con
+  = Just ( n', DataConI n full_con_ty ty_name
+#if __GLASGOW_HASKELL__ < 800
+                        fixity
 #endif
+         )
 
-  | Just (n', ty) <- findRecSelector n cons
+  | Just (n', rec_sel_info) <- findRecSelector n cons
+  , let (tvbs, sel_ty, con_res_ty) = extract_rec_sel_info rec_sel_info
+        -- See Note [Use unSigType in maybeReifyCon]
+        full_sel_ty = unSigType $ maybeForallT tvbs [] $ mkArrows [con_res_ty] sel_ty
       -- we don't try to ferret out naughty record selectors.
-#if __GLASGOW_HASKELL__ > 710
-  = Just (n', VarI n (maybeForallT tvbs [] $ mkArrows [result_ty] ty) Nothing)
-#else
-  = Just (n', VarI n (maybeForallT tvbs [] $ mkArrows [result_ty] ty) Nothing fixity)
+  = Just ( n', VarI n full_sel_ty Nothing
+#if __GLASGOW_HASKELL__ < 800
+                    fixity
 #endif
+         )
   where
-    result_ty = applyType (ConT ty_name) (map unSigTypeArg ty_args)
-      -- Make sure to call unSigTypeArg here. Otherwise, if you have this:
-      --
-      --   data D (a :: k) = MkD { unD :: Proxy a }
-      --
-      -- Then the type of unD will be reified as:
-      --
-      --   unD :: forall k (a :: k). D (a :: k) -> Proxy a
-      --
-      -- This is contrast to GHC's own reification, which will produce `D a`
-      -- (without the explicit kind signature) as the type of the first argument.
+    extract_rec_sel_info :: RecSelInfo -> ([TyVarBndrUnit], Type, Type)
+      -- Returns ( Selector type variable binders
+      --         , Record field type
+      --         , constructor result type )
+    extract_rec_sel_info rec_sel_info =
+      case rec_sel_info of
+        RecSelH98 sel_ty -> (h98_tvbs, sel_ty, h98_res_ty)
+        RecSelGADT sel_ty con_res_ty ->
+          ( freeVariablesWellScoped [con_res_ty, sel_ty]
+          , sel_ty, con_res_ty)
 
-    con_to_type (NormalC _ stys) = mkArrows (map snd    stys)  result_ty
-    con_to_type (RecC _ vstys)   = mkArrows (map thdOf3 vstys) result_ty
-    con_to_type (InfixC t1 _ t2) = mkArrows (map snd [t1, t2]) result_ty
-    con_to_type (ForallC bndrs cxt c) = ForallT bndrs cxt (con_to_type c)
-#if __GLASGOW_HASKELL__ > 710
-    con_to_type (GadtC _ stys rty)     = mkArrows (map snd    stys)  rty
-    con_to_type (RecGadtC _ vstys rty) = mkArrows (map thdOf3 vstys) rty
+    h98_tvbs   = freeVariablesWellScoped $ map probablyWrongUnTypeArg ty_args
+    h98_res_ty = applyType (ConT ty_name) ty_args
+
+#if __GLASGOW_HASKELL__ < 800
+    fixity = fromMaybe defaultFixity $ reifyFixityInDecs n _decs
 #endif
-#if __GLASGOW_HASKELL__ < 711
-    fixity = fromMaybe defaultFixity $ reifyFixityInDecs n decs
-#endif
-    tvbs = freeVariablesWellScoped $ map probablyWrongUnTypeArg ty_args
 maybeReifyCon _ _ _ _ _ = Nothing
 
+{-
+Note [Use unSigType in maybeReifyCon]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Make sure to call unSigType on the type of a reified data constructor or
+record selector. Otherwise, if you have this:
+
+  data D (a :: k) = MkD { unD :: Proxy a }
+
+Then the type of unD will be reified as:
+
+  unD :: forall k (a :: k). D (a :: k) -> Proxy a
+
+This is contrast to GHC's own reification, which will produce `D a`
+(without the explicit kind signature) as the type of the first argument.
+-}
+
+-- Reverse-engineer the type of a data constructor.
+con_to_type :: [TyVarBndrUnit] -- The type variables bound by a data type head.
+                               -- Only used for Haskell98-style constructors.
+            -> Type            -- The constructor result type.
+                               -- Only used for Haskell98-style constructors.
+            -> Con -> Type
+con_to_type h98_tvbs h98_result_ty con =
+  case go con of
+    (is_gadt, ty) | is_gadt   -> ty
+                  | otherwise -> maybeForallT h98_tvbs [] ty
+  where
+    -- Note that we deliberately ignore linear types and use (->) everywhere.
+    -- See [Gracefully handling linear types] in L.H.TH.Desugar.Core.
+    go :: Con -> (Bool, Type) -- The Bool is True when dealing with a GADT
+    go (NormalC _ stys)       = (False, mkArrows (map snd    stys)  h98_result_ty)
+    go (RecC _ vstys)         = (False, mkArrows (map thdOf3 vstys) h98_result_ty)
+    go (InfixC t1 _ t2)       = (False, mkArrows (map snd [t1, t2]) h98_result_ty)
+    go (ForallC bndrs cxt c)  = liftSnd (ForallT bndrs cxt) (go c)
+#if __GLASGOW_HASKELL__ > 710
+    go (GadtC _ stys rty)     = (True, mkArrows (map snd    stys)  rty)
+    go (RecGadtC _ vstys rty) = (True, mkArrows (map thdOf3 vstys) rty)
+#endif
+
 mkVarI :: Name -> [Dec] -> Info
 mkVarI n decs = mkVarITy n decs (maybe (no_type n) snd $ findType n decs)
 
@@ -449,14 +523,14 @@ findInstances n = map stripInstanceDec . concatMap match_instance
 #if __GLASGOW_HASKELL__ >= 807
     -- See Note [Rejigging reified type family equations variable binders]
     -- for why this is necessary.
-    rejig_tvbs :: [Type] -> Maybe [TyVarBndr]
+    rejig_tvbs :: [Type] -> Maybe [TyVarBndrUnit]
     rejig_tvbs ts =
       let tvbs = freeVariablesWellScoped ts
       in if null tvbs
          then Nothing
          else Just tvbs
 
-    rejig_data_inst_tvbs :: Cxt -> Type -> Maybe Kind -> Maybe [TyVarBndr]
+    rejig_data_inst_tvbs :: Cxt -> Type -> Maybe Kind -> Maybe [TyVarBndrUnit]
     rejig_data_inst_tvbs cxt lhs mk =
       rejig_tvbs $ cxt ++ [lhs] ++ maybeToList mk
 #endif
@@ -531,6 +605,7 @@ quantifyClassDecMethods (ClassD cxt cls_name cls_tvbs fds sub_decs)
     go (SigD n ty) =
       Just $ SigD n
            $ quantifyClassMethodType cls_name cls_tvbs prepend_cls ty
+    go d@(TySynInstD {})      = Just d
 #if __GLASGOW_HASKELL__ > 710
     go d@(OpenTypeFamilyD {}) = Just d
     go d@(DataFamilyD {})     = Just d
@@ -565,17 +640,17 @@ quantifyClassDecMethods dec = dec
 -- a single class method, like `method`, then one needs the class context to
 -- appear in the reified type, so `True` is appropriate.
 quantifyClassMethodType
-  :: Name        -- ^ The class name.
-  -> [TyVarBndr] -- ^ The class's type variable binders.
-  -> Bool        -- ^ If 'True', prepend a class predicate.
-  -> Type        -- ^ The method type.
+  :: Name            -- ^ The class name.
+  -> [TyVarBndrUnit] -- ^ The class's type variable binders.
+  -> Bool            -- ^ If 'True', prepend a class predicate.
+  -> Type            -- ^ The method type.
   -> Type
 quantifyClassMethodType cls_name cls_tvbs prepend meth_ty =
   add_cls_cxt quantified_meth_ty
   where
     add_cls_cxt :: Type -> Type
     add_cls_cxt
-      | prepend   = ForallT all_cls_tvbs cls_cxt
+      | prepend   = ForallT (changeTVFlags SpecifiedSpec all_cls_tvbs) cls_cxt
       | otherwise = id
 
     cls_cxt :: Cxt
@@ -594,12 +669,13 @@ quantifyClassMethodType cls_name cls_tvbs prepend meth_ty =
       | otherwise
       = ForallT meth_tvbs [] meth_ty
 
-    meth_tvbs :: [TyVarBndr]
-    meth_tvbs = deleteFirstsBy ((==) `on` tvName)
+    meth_tvbs :: [TyVarBndrSpec]
+    meth_tvbs = changeTVFlags SpecifiedSpec $
+                deleteFirstsBy ((==) `on` tvName)
                   (freeVariablesWellScoped [meth_ty]) all_cls_tvbs
 
     -- Explicitly quantify any kind variables bound by the class, if any.
-    all_cls_tvbs :: [TyVarBndr]
+    all_cls_tvbs :: [TyVarBndrUnit]
     all_cls_tvbs = freeVariablesWellScoped $ map tvbToTypeWithSig cls_tvbs
 
 stripInstanceDec :: Dec -> Dec
@@ -614,11 +690,13 @@ mkArrows :: [Type] -> Type -> Type
 mkArrows []     res_ty = res_ty
 mkArrows (t:ts) res_ty = AppT (AppT ArrowT t) $ mkArrows ts res_ty
 
-maybeForallT :: [TyVarBndr] -> Cxt -> Type -> Type
+maybeForallT :: [TyVarBndrUnit] -> Cxt -> Type -> Type
 maybeForallT tvbs cxt ty
   | null tvbs && null cxt        = ty
-  | ForallT tvbs2 cxt2 ty2 <- ty = ForallT (tvbs ++ tvbs2) (cxt ++ cxt2) ty2
-  | otherwise                    = ForallT tvbs cxt ty
+  | ForallT tvbs2 cxt2 ty2 <- ty = ForallT (tvbs_spec ++ tvbs2) (cxt ++ cxt2) ty2
+  | otherwise                    = ForallT tvbs_spec cxt ty
+  where
+    tvbs_spec = changeTVFlags SpecifiedSpec tvbs
 
 findCon :: Name -> [Con] -> Maybe (Named Con)
 findCon n = firstMatch match_con
@@ -645,18 +723,27 @@ findCon n = firstMatch match_con
                           Nothing -> Nothing
 #endif
 
-findRecSelector :: Name -> [Con] -> Maybe (Named Type)
+data RecSelInfo
+  = RecSelH98  Type -- The record field's type
+  | RecSelGADT Type -- The record field's type
+               Type -- The GADT return type
+
+findRecSelector :: Name -> [Con] -> Maybe (Named RecSelInfo)
 findRecSelector n = firstMatch match_con
   where
-    match_con (RecC _ vstys)       = firstMatch match_rec_sel vstys
+    match_con :: Con -> Maybe (Named RecSelInfo)
+    match_con (RecC _ vstys)            = fmap (liftSnd RecSelH98) $
+                                          firstMatch match_rec_sel vstys
 #if __GLASGOW_HASKELL__ >= 800
-    match_con (RecGadtC _ vstys _) = firstMatch match_rec_sel vstys
+    match_con (RecGadtC _ vstys ret_ty) = fmap (liftSnd (`RecSelGADT` ret_ty)) $
+                                          firstMatch match_rec_sel vstys
 #endif
-    match_con (ForallC _ _ c)      = match_con c
-    match_con _                    = Nothing
+    match_con (ForallC _ _ c)           = match_con c
+    match_con _                         = Nothing
 
-    match_rec_sel (n', _, ty) | n `nameMatches` n' = Just (n', ty)
-    match_rec_sel _                                = Nothing
+    match_rec_sel (n', _, sel_ty)
+      | n `nameMatches` n' = Just (n', sel_ty)
+    match_rec_sel _        = Nothing
 
 ---------------------------------
 -- Reifying fixities
@@ -684,8 +771,8 @@ reifyFixity :: Name -> Q (Maybe Fixity)
 reifyFixity = qReifyFixity
 #endif
 
--- | Like 'reifyWithLocals_maybe', but for fixities. Note that a return of
--- @Nothing@ might mean that the name is not in scope, or it might mean
+-- | Like 'reifyWithLocals_maybe', but for fixities. Note that a return value
+-- of @Nothing@ might mean that the name is not in scope, or it might mean
 -- that the name has no assigned fixity. (Use 'reifyWithLocals_maybe' if
 -- you really need to tell the difference.)
 reifyFixityWithLocals :: DsMonad q => Name -> q (Maybe Fixity)
@@ -694,7 +781,358 @@ reifyFixityWithLocals name = qRecover
   (qReifyFixity name)
 
 --------------------------------------
--- Lookuping name value and type names
+-- Reifying types
+--------------------------------------
+--
+-- This section allows GHC <8.9 to call reifyFixity
+
+#if __GLASGOW_HASKELL__ < 809
+qReifyType :: forall m. Quasi m => Name -> m Type
+qReifyType name = do
+  info <- qReify name
+  case infoType info <|> info_kind info of
+    Just t  -> return t
+    Nothing -> fail $ "Could not reify the full type of " ++ nameBase name
+  where
+    info_kind :: Info -> Maybe Kind
+    info_kind info = do
+      dec <- case info of
+               ClassI d _  -> Just d
+               TyConI d    -> Just d
+               FamilyI d _ -> Just d
+               _           -> Nothing
+      match_cusk name dec
+
+{- | @reifyType nm@ attempts to find the type or kind of @nm@. For example,
+@reifyType 'not@   returns @Bool -> Bool@, and
+@reifyType ''Bool@ returns @Type@.
+This works even if there's no explicit signature and the type or kind is inferred.
+-}
+reifyType :: Name -> Q Type
+reifyType = qReifyType
+#endif
+
+-- | Like 'reifyTypeWithLocals_maybe', but throws an exception upon failure,
+-- warning the user about separating splices.
+reifyTypeWithLocals :: DsMonad q => Name -> q Type
+reifyTypeWithLocals name = do
+  m_info <- reifyTypeWithLocals_maybe name
+  case m_info of
+    Nothing -> reifyFail name
+    Just i  -> return i
+
+-- | Like 'reifyWithLocals_maybe' but for types and kinds. Note that a return
+-- value of @Nothing@ might mean that the name is not in scope, or it might
+-- mean that the full type of the name cannot be determined. (Use
+-- 'reifyWithLocals_maybe' if you really need to tell the difference.)
+reifyTypeWithLocals_maybe :: DsMonad q => Name -> q (Maybe Type)
+reifyTypeWithLocals_maybe name = do
+#if __GLASGOW_HASKELL__ >= 809
+  cusks <- qIsExtEnabled CUSKs
+#else
+  -- On earlier GHCs, the behavior of -XCUSKs was the norm.
+  let cusks = True
+#endif
+  qRecover (return . reifyTypeInDecs cusks name =<< localDeclarations)
+           (Just `fmap` qReifyType name)
+
+-- | Look through a list of declarations and return its full type, if
+-- available.
+reifyTypeInDecs :: Bool -> Name -> [Dec] -> Maybe Type
+reifyTypeInDecs cusks name decs =
+  (reifyInDecs name decs >>= infoType) <|> findKind cusks name decs
+
+-- Extract the type information (if any) contained in an Info.
+infoType :: Info -> Maybe Type
+infoType info =
+  case info of
+    ClassOpI _ t _
+#if __GLASGOW_HASKELL__ < 800
+             _
+#endif
+                   -> Just t
+    DataConI _ t _
+#if __GLASGOW_HASKELL__ < 800
+             _
+#endif
+                   -> Just t
+    VarI _ t _
+#if __GLASGOW_HASKELL__ < 800
+         _
+#endif
+                   -> Just t
+    TyVarI _ t     -> Just t
+#if __GLASGOW_HASKELL__ >= 802
+    PatSynI _ t    -> Just t
+#endif
+    _              -> Nothing
+
+-- Like findType, but instead searching for kind signatures.
+-- This mostly searches through `KiSigD`s, but if the -XCUSKs extension is
+-- enabled, this also retrieves kinds for declarations with CUSKs.
+findKind :: Bool -- Is -XCUSKs enabled?
+         -> Name -> [Dec] -> Maybe Kind
+findKind cusks name decls =
+      firstMatch (match_kind_sig name decls) decls
+  <|> whenAlt cusks (firstMatch (match_cusk name) decls)
+
+-- Look for a declaration's kind by searching for its standalone kind
+-- signature, if available.
+match_kind_sig :: Name -> [Dec] -> Dec -> Maybe Kind
+match_kind_sig n decs (ClassD _ n' tvbs _ sub_decs)
+  -- If a class has a standalone kind signature, then we can determine the
+  -- full kind of its associated types in 99% of cases.
+  -- See Note [The limitations of standalone kind signatures] for what
+  -- happens in the other 1% of cases.
+  | Just ki <- firstMatch (find_kind_sig n') decs
+  , let (arg_kis, _res_ki) = unravelType ki
+        mb_vis_arg_kis     = map vis_arg_kind_maybe $ filterVisFunArgs arg_kis
+        cls_tvb_kind_map   =
+          Map.fromList [ (tvName tvb, tvb_kind)
+                       | (tvb, mb_vis_arg_ki) <- zip tvbs mb_vis_arg_kis
+                       , Just tvb_kind <- [mb_vis_arg_ki <|> tvb_kind_maybe tvb]
+                       ]
+  = firstMatch (find_assoc_type_kind n cls_tvb_kind_map) sub_decs
+match_kind_sig n _ dec = find_kind_sig n dec
+
+find_kind_sig :: Name -> Dec -> Maybe Kind
+#if __GLASGOW_HASKELL__ >= 809
+find_kind_sig n (KiSigD n' ki)
+  | n `nameMatches` n' = Just ki
+#endif
+find_kind_sig _ _ = Nothing
+
+-- Compute a declaration's kind by retrieving its CUSK, if it has one.
+-- This is only done when -XCUSKs is enabled, or on older GHCs where
+-- CUSKs were the only means of specifying this information.
+match_cusk :: Name -> Dec -> Maybe Kind
+#if __GLASGOW_HASKELL__ >= 800
+match_cusk n (DataD _ n' tvbs m_ki _ _)
+  | n `nameMatches` n'
+  = datatype_kind tvbs m_ki
+match_cusk n (NewtypeD _ n' tvbs m_ki _ _)
+  | n `nameMatches` n'
+  = datatype_kind tvbs m_ki
+match_cusk n (DataFamilyD n' tvbs m_ki)
+  | n `nameMatches` n'
+  = open_ty_fam_kind tvbs m_ki
+match_cusk n (OpenTypeFamilyD (TypeFamilyHead n' tvbs res_sig _))
+  | n `nameMatches` n'
+  = open_ty_fam_kind tvbs (res_sig_to_kind res_sig)
+match_cusk n (ClosedTypeFamilyD (TypeFamilyHead n' tvbs res_sig _) _)
+  | n `nameMatches` n'
+  = closed_ty_fam_kind tvbs (res_sig_to_kind res_sig)
+#else
+match_cusk n (DataD _ n' tvbs _ _)
+  | n `nameMatches` n'
+  = datatype_kind tvbs Nothing
+match_cusk n (NewtypeD _ n' tvbs _ _)
+  | n `nameMatches` n'
+  = datatype_kind tvbs Nothing
+match_cusk n (FamilyD _ n' tvbs m_ki)
+  | n `nameMatches` n'
+  = open_ty_fam_kind tvbs m_ki
+match_cusk n (ClosedTypeFamilyD n' tvbs m_ki _)
+  | n `nameMatches` n'
+  = closed_ty_fam_kind tvbs m_ki
+#endif
+match_cusk n (TySynD n' tvbs rhs)
+  | n `nameMatches` n'
+  = ty_syn_kind tvbs rhs
+match_cusk n (ClassD _ n' tvbs _ sub_decs)
+  | n `nameMatches` n'
+  = class_kind tvbs
+  | -- An associated type family can only have a CUSK if its parent class
+    -- also has a CUSK.
+    all tvb_is_kinded tvbs
+  , let cls_tvb_kind_map = Map.fromList [ (tvName tvb, tvb_kind)
+                                        | tvb <- tvbs
+                                        , Just tvb_kind <- [tvb_kind_maybe tvb]
+                                        ]
+  = firstMatch (find_assoc_type_kind n cls_tvb_kind_map) sub_decs
+match_cusk _ _ = Nothing
+
+-- Uncover the kind of an associated type family. There is an invariant
+-- that this function should only ever be called when the kind of the
+-- parent class is known (i.e., if it has a standalone kind signature or a
+-- CUSK). Despite this, it is possible for this function to return Nothing.
+-- See Note [The limitations of standalone kind signatures].
+find_assoc_type_kind :: Name -> Map Name Kind -> Dec -> Maybe Kind
+find_assoc_type_kind n cls_tvb_kind_map sub_dec =
+  case sub_dec of
+#if __GLASGOW_HASKELL__ >= 800
+    DataFamilyD n' tf_tvbs m_ki
+      |  n `nameMatches` n'
+      -> build_kind (map ascribe_tf_tvb_kind tf_tvbs) (default_res_ki m_ki)
+    OpenTypeFamilyD (TypeFamilyHead n' tf_tvbs res_sig _)
+      |  n `nameMatches` n'
+      -> build_kind (map ascribe_tf_tvb_kind tf_tvbs)
+                    (default_res_ki $ res_sig_to_kind res_sig)
+#else
+    FamilyD _ n' tf_tvbs m_ki
+      |  n `nameMatches` n'
+      -> build_kind (map ascribe_tf_tvb_kind tf_tvbs) (default_res_ki m_ki)
+#endif
+    _ -> Nothing
+  where
+    ascribe_tf_tvb_kind :: TyVarBndrUnit -> TyVarBndrUnit
+    ascribe_tf_tvb_kind tvb =
+      elimTV (\tvn -> kindedTV tvn $ fromMaybe StarT $ Map.lookup tvn cls_tvb_kind_map)
+             (\_ _ -> tvb)
+             tvb
+
+-- Data types have CUSKs when:
+--
+-- 1. All of their type variables have explicit kinds.
+-- 2. All kind variables in the result kind are explicitly quantified.
+datatype_kind :: [TyVarBndrUnit] -> Maybe Kind -> Maybe Kind
+datatype_kind tvbs m_ki =
+  whenAlt (all tvb_is_kinded tvbs && ki_fvs_are_bound) $
+  build_kind tvbs (default_res_ki m_ki)
+  where
+    ki_fvs_are_bound :: Bool
+    ki_fvs_are_bound =
+      let ki_fvs   = Set.fromList $ foldMap freeVariables m_ki
+          tvb_vars = Set.fromList $ freeVariables $ map tvbToTypeWithSig tvbs
+      in ki_fvs `Set.isSubsetOf` tvb_vars
+
+-- Classes have CUSKs when all of their type variables have explicit kinds.
+class_kind :: [TyVarBndrUnit] -> Maybe Kind
+class_kind tvbs = whenAlt (all tvb_is_kinded tvbs) $
+                  build_kind tvbs ConstraintT
+
+-- Open type families and data families always have CUSKs. Type variables
+-- without explicit kinds default to Type, as does the return kind if it
+-- is not specified.
+open_ty_fam_kind :: [TyVarBndrUnit] -> Maybe Kind -> Maybe Kind
+open_ty_fam_kind tvbs m_ki =
+  build_kind (map default_tvb tvbs) (default_res_ki m_ki)
+
+-- Closed type families have CUSKs when:
+--
+-- 1. All of their type variables have explicit kinds.
+-- 2. An explicit return kind is supplied.
+closed_ty_fam_kind :: [TyVarBndrUnit] -> Maybe Kind -> Maybe Kind
+closed_ty_fam_kind tvbs m_ki =
+  case m_ki of
+    Just ki -> whenAlt (all tvb_is_kinded tvbs) $
+               build_kind tvbs ki
+    Nothing -> Nothing
+
+-- Type synonyms have CUSKs when:
+--
+-- 1. All of their type variables have explicit kinds.
+-- 2. The right-hand-side type is annotated with an explicit kind.
+ty_syn_kind :: [TyVarBndrUnit] -> Type -> Maybe Kind
+ty_syn_kind tvbs rhs =
+  case rhs of
+    SigT _ ki -> whenAlt (all tvb_is_kinded tvbs) $
+                 build_kind tvbs ki
+    _         -> Nothing
+
+-- Attempt to construct the full kind of a type-level declaration from its
+-- type variable binders and return kind. Do note that the result type of
+-- this function is `Maybe Kind` because there are situations where even
+-- this amount of information is not sufficient to determine the full kind.
+-- See Note [The limitations of standalone kind signatures].
+build_kind :: [TyVarBndrUnit] -> Kind -> Maybe Kind
+build_kind arg_kinds res_kind =
+  fmap quantifyType $ fst $
+  foldr go (Just res_kind, Set.fromList (freeVariables res_kind)) arg_kinds
+  where
+    go :: TyVarBndrUnit -> (Maybe Kind, Set Name) -> (Maybe Kind, Set Name)
+    go tvb (res, res_fvs) =
+      elimTV (\n ->
+               ( if n `Set.member` res_fvs
+                 then forall_vis tvb res
+                 else Nothing -- We have a type variable binder without an
+                              -- explicit kind that is not used dependently, so
+                              -- we cannot build a kind from it. This is the
+                              -- only case where we return Nothing.
+               , res_fvs
+               ))
+             (\n k ->
+               ( if n `Set.member` res_fvs
+                 then forall_vis tvb res
+                 else fmap (ArrowT `AppT` k `AppT`) res
+               , Set.fromList (freeVariables k) `Set.union` res_fvs
+               ))
+             tvb
+
+    forall_vis :: TyVarBndrUnit -> Maybe Kind -> Maybe Kind
+#if __GLASGOW_HASKELL__ >= 809
+    forall_vis tvb m_ki = fmap (ForallVisT [tvb]) m_ki
+      -- One downside of this approach is that we generate kinds like this:
+      --
+      --   forall a -> forall b -> forall c -> (a, b, c)
+      --
+      -- Instead of this more compact kind:
+      --
+      --   forall a b c -> (a, b, c)
+      --
+      -- Thankfully, the difference is only cosmetic.
+#else
+    forall_vis _   _    = Nothing
+#endif
+
+tvb_is_kinded :: TyVarBndr_ flag -> Bool
+tvb_is_kinded = isJust . tvb_kind_maybe
+
+tvb_kind_maybe :: TyVarBndr_ flag -> Maybe Kind
+tvb_kind_maybe = elimTV (\_ -> Nothing) (\_ k -> Just k)
+
+vis_arg_kind_maybe :: VisFunArg -> Maybe Kind
+vis_arg_kind_maybe (VisFADep tvb) = tvb_kind_maybe tvb
+vis_arg_kind_maybe (VisFAAnon k)  = Just k
+
+default_tvb :: TyVarBndrUnit -> TyVarBndrUnit
+default_tvb tvb = elimTV (\n -> kindedTV n StarT) (\_ _ -> tvb) tvb
+
+default_res_ki :: Maybe Kind -> Kind
+default_res_ki = fromMaybe StarT
+
+#if __GLASGOW_HASKELL__ >= 800
+res_sig_to_kind :: FamilyResultSig -> Maybe Kind
+res_sig_to_kind NoSig          = Nothing
+res_sig_to_kind (KindSig k)    = Just k
+res_sig_to_kind (TyVarSig tvb) = tvb_kind_maybe tvb
+#endif
+
+whenAlt :: Alternative f => Bool -> f a -> f a
+whenAlt b fa = if b then fa else empty
+
+{-
+Note [The limitations of standalone kind signatures]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+A current limitation of StandaloneKindSignatures is that they cannot be applied
+to associated type families. This can have some surprising consequences.
+Consider the following code, taken from
+https://gitlab.haskell.org/ghc/ghc/issues/17072#note_221324:
+
+  type C :: forall a -> a -> Constraint
+  class C a b where
+    type T a :: Type
+
+The parent class C has a standalone kind signature, so GHC treats its
+associated types as if they had CUSKs. Can th-desugar figure out the kind
+that GHC gives to T?
+
+Unfortunately, the answer is "not easily". This is because `type T a` says
+nothing about the kind of `a`, so th-desugar's only other option is to inspect
+the kind signature for C. Even this is for naught, as the `forall a -> ...`
+part doesn't state the kind of `a` either! The only way to know that the kind
+of `a` should be Type is to infer that from the rest of the kind
+(`a -> Constraint`), but this gets perilously close to requiring full kind
+inference, which is rather unwieldy in Template Haskell.
+
+In cases like T, we simply give up and return Nothing when trying to reify
+its kind. It's not ideal, but them's the breaks when you try to extract kinds
+from syntax. There is a rather simple workaround available: just write
+`type C :: forall (a :: Type) -> a -> Constraint` instead.
+-}
+
+--------------------------------------
+-- Looking up name value and type names
 --------------------------------------
 
 -- | Like 'lookupValueName' from Template Haskell, but looks also in 'Names' of
