@@ -8,7 +8,8 @@ rae@cs.brynmawr.edu
              RankNTypes, TypeFamilies,
              DataKinds, ConstraintKinds, PolyKinds, MultiParamTypeClasses,
              FlexibleInstances, ExistentialQuantification,
-             ScopedTypeVariables, GADTs, ViewPatterns, TupleSections #-}
+             ScopedTypeVariables, GADTs, ViewPatterns, TupleSections,
+             TypeOperators #-}
 {-# OPTIONS -fno-warn-incomplete-patterns -fno-warn-overlapping-patterns
             -fno-warn-unused-matches -fno-warn-type-defaults
             -fno-warn-missing-signatures -fno-warn-unused-do-bind
@@ -26,6 +27,10 @@ rae@cs.brynmawr.edu
 {-# LANGUAGE QuantifiedConstraints #-}
 #endif
 
+#if __GLASGOW_HASKELL__ >= 809
+{-# LANGUAGE StandaloneKindSignatures #-}
+#endif
+
 module Main where
 
 import Prelude hiding ( exp )
@@ -38,6 +43,8 @@ import Splices
 import qualified DsDec
 import qualified Dec
 import Dec ( RecordSel )
+import ReifyTypeCUSKs
+import ReifyTypeSigs
 import Language.Haskell.TH.Desugar
 import qualified Language.Haskell.TH.Desugar.OSet as OS
 import Language.Haskell.TH.Desugar.Expand  ( expandUnsoundly )
@@ -49,10 +56,13 @@ import Control.Monad
 import Control.Applicative
 #endif
 
-import Data.Function ( on )
 import qualified Data.Map as M
 import Data.Proxy
 
+#if __GLASGOW_HASKELL__ >= 900
+import Prelude as P
+#endif
+
 -- |
 -- Convert a HUnit test suite to a spec.  This can be used to run existing
 -- HUnit tests with Hspec.
@@ -137,6 +147,9 @@ tests = test [ "sections" ~: $test1_sections  @=? $(dsSplice test1_sections)
 #endif
 #if __GLASGOW_HASKELL__ >= 809
              , "tuple_sections"  ~: $test51_tuple_sections  @=? $(dsSplice test51_tuple_sections)
+#endif
+#if __GLASGOW_HASKELL__ >= 900
+             , "qual_do"         ~: $test52_qual_do         @=? $(dsSplice test52_qual_do)
 #endif
              ]
 
@@ -223,10 +236,10 @@ test_bug8884 = $(do info <- reify ''Poly
 
 flatten_dvald :: Bool
 flatten_dvald = let s1 = $(flatten_dvald_test)
-                    s2 = $(do exp <- flatten_dvald_test
-                              DLetE ddecs dexp <- dsExp exp
+                    s2 = $(do expr <- flatten_dvald_test
+                              DLetE ddecs dexpr <- dsExp expr
                               flattened <- fmap concat $ mapM flattenDValD ddecs
-                              return $ expToTH $ DLetE flattened dexp ) in
+                              return $ expToTH $ DLetE flattened dexpr ) in
                 s1 == s2
 
 test_rec_sels :: Bool
@@ -268,7 +281,7 @@ test_stuck_tyfam_expansion =
                    (decsToTH [ -- type family F (x :: k) :: k
                                DOpenTypeFamilyD
                                  (DTypeFamilyHead fam_name
-                                                  [DKindedTV x (DVarT k)]
+                                                  [DKindedTV x () (DVarT k)]
                                                   (DKindSig (DVarT k))
                                                   Nothing)
                                -- type instance F (x :: ()) = x
@@ -294,17 +307,23 @@ test_t92 :: Bool
 test_t92 =
   $(do a <- newName "a"
        f <- newName "f"
-       let t = DForallT [DPlainTV f] [] (DVarT f `DAppT` DVarT a)
-       toposortTyVarsOf [t] `eqTHSplice` [DPlainTV a])
+       let t = DForallT (DForallInvis [DPlainTV f SpecifiedSpec])
+                        (DVarT f `DAppT` DVarT a)
+       toposortTyVarsOf [t] `eqTHSplice` [DPlainTV a ()])
 
 test_t97 :: Bool
 test_t97 =
   $(do a <- newName "a"
        k <- newName "k"
        let orig_ty = DForallT
-                       [DKindedTV a (DConT ''Constant `DAppT` DConT ''Int `DAppT` DVarT k)]
-                       [] (DVarT a)
-           expected_ty = DForallT [DKindedTV a (DVarT k)] [] (DVarT a)
+                       (DForallInvis
+                         [DKindedTV a SpecifiedSpec
+                                      (DConT ''Constant `DAppT` DConT ''Int
+                                                        `DAppT` DVarT k)])
+                       (DVarT a)
+           expected_ty = DForallT (DForallInvis
+                                    [DKindedTV a SpecifiedSpec (DVarT k)])
+                                  (DVarT a)
        expanded_ty <- expandType orig_ty
        expected_ty `eqTHSplice` expanded_ty)
 
@@ -317,21 +336,50 @@ test_getDataD_kind_sig =
                 data_kind_sig = DArrowT `DAppT` type_kind `DAppT`
                                   (DArrowT `DAppT` type_kind `DAppT` type_kind)
             (tvbs, _) <- withLocalDeclarations
-                           (decToTH (DDataD Data [] data_name [DPlainTV a]
-                                            (Just data_kind_sig) [] []))
+                           [decToTH (DDataD Data [] data_name [DPlainTV a ()]
+                                            (Just data_kind_sig) [] [])]
                            (getDataD "th-desugar: Impossible" data_name)
             [| $(Syn.lift (length tvbs)) |])
 #else
   True -- DataD didn't have the ability to store kind signatures prior to GHC 8.0
 #endif
 
-test_t102 :: Bool
+test_t100 :: Bool
+test_t100 =
+#if __GLASGOW_HASKELL__ >= 800
+  $(do decs <- [d| data T b where
+                     MkT :: forall a. { unT :: a } -> T a |]
+       info <- withLocalDeclarations decs (dsReify (mkName "unT"))
+       let -- forall a. T a -> a
+           exp_ty = DForallT (DForallInvis [DPlainTV (mkName "a") SpecifiedSpec]) $
+                    DArrowT `DAppT` (DConT (mkName "T") `DAppT` DVarT (mkName "a"))
+                            `DAppT` DVarT (mkName "a")
+       case info of
+         Just (DVarI _ actual_ty _) -> exp_ty `eqTHSplice` actual_ty
+         _                          -> [| False |])
+#else
+  True -- RecGadtC didn't exist prior to GHC 8.0, do the quote above will
+       -- normalize to `data T b = MkT { unT :: b }`. This defeats the point of
+       -- this test, and to make things worse, this will cause `dsReify` to
+       -- return a different type for unT (forall b. T b -> b). Let's just not
+       -- bother testing this on pre-8.0 GHCs.
+#endif
+
+test_t102 :: [Bool]
 test_t102 =
-  $(do decs <- [d| data Foo x where MkFoo :: forall a. { unFoo :: a } -> Foo a |]
-       withLocalDeclarations decs $ do
-         [DDataD _ _ foo [DPlainTV x] _ cons _] <- dsDecs decs
-         recs <- getRecordSelectors (DConT foo `DAppT` DVarT x) cons
-         (length recs `div` 2) `eqTHSplice` 1)
+  $(do decs1 <- [d| data Foo x where MkFoo :: forall a. { unFoo :: a } -> Foo a |]
+       let b1 = withLocalDeclarations decs1 $ do
+                  [DDataD _ _ _ _ _ cons1 _] <- dsDecs decs1
+                  recs1 <- getRecordSelectors cons1
+                  (length recs1 `div` 2) `eqTHSplice` 1
+       decs2 <- [d| data HList l where
+                      Nil  :: HList '[]
+                      (:>) :: { hhead :: x, htail :: HList xs } -> HList (x ': xs) |]
+       let b2 = withLocalDeclarations decs2 $ do
+                  [DDataD _ _ _ _ _ cons2 _] <- dsDecs decs2
+                  recs2 <- getRecordSelectors cons2
+                  (length recs2 `div` 2) `eqTHSplice` 2
+       [| [$b1, $b2] |])
 
 test_t103 :: Bool
 test_t103 =
@@ -339,7 +387,7 @@ test_t103 =
   $(do decs <- [d| data P (a :: k) = MkP |]
        [DDataD _ _ _ _ _ [DCon tvbs _ _ _ _] _] <- dsDecs decs
        case tvbs of
-         [DPlainTV k, DKindedTV a (DVarT k')]
+         [DPlainTV k SpecifiedSpec, DKindedTV a SpecifiedSpec (DVarT k')]
            |  nameBase k == "k"
            ,  nameBase a == "a"
            ,  k == k'
@@ -354,9 +402,12 @@ test_t112 :: [Bool]
 test_t112 =
   $(do a <- newName "a"
        b <- newName "b"
-       let [aVar, bVar] = map DVarT    [a, b]
-           [aTvb, bTvb] = map DPlainTV [a, b]
-       let fvsABExpected = [aTvb, bTvb]
+       let aVar = DVarT a
+           bVar = DVarT b
+           aTvb = DPlainTV a ()
+           bTvb = DPlainTV b ()
+
+           fvsABExpected = [aTvb, bTvb]
            fvsABActual   = toposortTyVarsOf [aVar, bVar]
 
            fvsBAExpected = [bTvb, aTvb]
@@ -366,16 +417,37 @@ test_t112 =
            eqBA = fvsBAExpected `eqTH` fvsBAActual
        [| [eqAB, eqBA] |])
 
+test_t132 :: Bool
+test_t132 =
+  $(do let c      = mkName "C"
+           m      = mkName "m"
+           a      = mkName "a"
+           fixity = Fixity 5 InfixR
+           -- Defines a class with a fixity declaration inside, i.e.,
+           --
+           --   class C a where
+           --     infixr 5 `m`
+           --     m :: a
+           --
+           -- We define this by hand to avoid GHC#17608 on pre-9.0 GHCs.
+           decs = sweeten [ DClassD [] c [DPlainTV a ()] []
+                            [ DLetDec (DInfixD fixity m)
+                            , DLetDec (DSigD m (DVarT a))
+                            ]
+                          ]
+           expected = Just fixity
+       actual <- withLocalDeclarations decs (reifyFixityWithLocals m)
+       expected `eqTHSplice` actual)
+
 -- Unit tests for functions that compute free variables (e.g., fvDType)
 test_fvs :: [Bool]
 test_fvs =
   $(do a <- newName "a"
 
        let -- (Show a => Show (Maybe a)) => String
-           ty1 = DForallT
-                   []
-                   [DForallT [] [DConT ''Show `DAppT` DVarT a]
-                                (DConT ''Show `DAppT` (DConT ''Maybe `DAppT` DVarT a))]
+           ty1 = DConstrainedT
+                   [DConstrainedT [DConT ''Show `DAppT` DVarT a]
+                                  (DConT ''Show `DAppT` (DConT ''Maybe `DAppT` DVarT a))]
                    (DConT ''String)
            b1 = fvDType ty1 `eqTH` OS.singleton a -- #93
 
@@ -392,15 +464,21 @@ test_kind_substitution =
                  -- (Nothing :: Maybe a)
            ty1 = DSigT (DConT 'Nothing) (DConT ''Maybe `DAppT` DVarT a)
                  -- forall (c :: a). c
-           ty2 = DForallT [DKindedTV c (DVarT a)] [] (DVarT c)
+           ty2 = DForallT (DForallInvis [DKindedTV c SpecifiedSpec (DVarT a)])
+                          (DVarT c)
                  -- forall a (c :: a). c
-           ty3 = DForallT [DPlainTV a, DKindedTV c (DVarT a)] [] (DVarT c)
+           ty3 = DForallT (DForallInvis [ DPlainTV  a SpecifiedSpec
+                                        , DKindedTV c SpecifiedSpec (DVarT a)
+                                        ])
+                          (DVarT c)
                  -- forall (a :: k) k (b :: k). Proxy b -> Proxy a
-           ty4 = DForallT [ DKindedTV a (DVarT k)
-                          , DPlainTV k
-                          , DKindedTV b (DVarT k)
-                          ] [] (DArrowT `DAppT` (DConT ''Proxy `DAppT` DVarT b)
-                                        `DAppT` (DConT ''Proxy `DAppT` DVarT a))
+           ty4 = DForallT (DForallInvis
+                             [ DKindedTV a SpecifiedSpec (DVarT k)
+                             , DPlainTV  k SpecifiedSpec
+                             , DKindedTV b SpecifiedSpec (DVarT k)
+                             ])
+                          (DArrowT `DAppT` (DConT ''Proxy `DAppT` DVarT b)
+                                   `DAppT` (DConT ''Proxy `DAppT` DVarT a))
 
        substTy1 <- substTy subst ty1
        substTy2 <- substTy subst ty2
@@ -460,7 +538,7 @@ $(return [])  -- somehow, this is necessary to get the staging correct for the
 normal_reifications :: [String]
 normal_reifications = $(do infos <- mapM reify reifyDecsNames
                            ListE <$> mapM (Syn.lift . show . Just)
-                                          (dropTrailing0s $ unqualify infos))
+                                          (dropTrailing0s $ delinearize $ unqualify infos))
 
 zipWith3M :: Monad m => (a -> b -> c -> m d) -> [a] -> [b] -> [c] -> m [d]
 zipWith3M f (a:as) (b:bs) (c:cs) = liftM2 (:) (f a b c) (zipWith3M f as bs cs)
@@ -486,33 +564,29 @@ test_roundtrip = $( do exprs <- sequence test_exprs
 
 test_matchTy :: [Bool]
 test_matchTy =
-  [ matchTy NoIgnore (DVarT a) (DConT ''Bool) `eq` Just (M.singleton a (DConT ''Bool))
-  , matchTy NoIgnore (DVarT a) (DVarT a) `eq` Just (M.singleton a (DVarT a))
-  , matchTy NoIgnore (DVarT a) (DVarT b) `eq` Just (M.singleton a (DVarT b))
+  [ matchTy NoIgnore (DVarT a) (DConT ''Bool) == Just (M.singleton a (DConT ''Bool))
+  , matchTy NoIgnore (DVarT a) (DVarT a) == Just (M.singleton a (DVarT a))
+  , matchTy NoIgnore (DVarT a) (DVarT b) == Just (M.singleton a (DVarT b))
   , matchTy NoIgnore (DConT ''Either `DAppT` DVarT a `DAppT` DVarT b)
                      (DConT ''Either `DAppT` DConT ''Int `DAppT` DConT ''Bool)
-    `eq` Just (M.fromList [(a, DConT ''Int), (b, DConT ''Bool)])
+    == Just (M.fromList [(a, DConT ''Int), (b, DConT ''Bool)])
   , matchTy NoIgnore (DConT ''Either `DAppT` DVarT a `DAppT` DVarT a)
                      (DConT ''Either `DAppT` DConT ''Int `DAppT` DConT ''Int)
-    `eq` Just (M.singleton a (DConT ''Int))
+    == Just (M.singleton a (DConT ''Int))
   , matchTy NoIgnore (DConT ''Either `DAppT` DVarT a `DAppT` DVarT a)
                      (DConT ''Either `DAppT` DConT ''Int `DAppT` DConT ''Bool)
-    `eq` Nothing
-  , matchTy NoIgnore (DConT ''Int) (DConT ''Bool) `eq` Nothing
-  , matchTy NoIgnore (DConT ''Int) (DConT ''Int) `eq` Just M.empty
-  , matchTy NoIgnore (DConT ''Int) (DVarT a) `eq` Nothing
-  , matchTy NoIgnore (DVarT a `DSigT` DConT ''Bool) (DConT ''Int) `eq` Nothing
+    == Nothing
+  , matchTy NoIgnore (DConT ''Int) (DConT ''Bool) == Nothing
+  , matchTy NoIgnore (DConT ''Int) (DConT ''Int) == Just M.empty
+  , matchTy NoIgnore (DConT ''Int) (DVarT a) == Nothing
+  , matchTy NoIgnore (DVarT a `DSigT` DConT ''Bool) (DConT ''Int) == Nothing
   , matchTy YesIgnore (DVarT a `DSigT` DConT ''Bool) (DConT ''Int)
-    `eq` Just (M.singleton a (DConT ''Int))
+    == Just (M.singleton a (DConT ''Int))
   ]
   where
     a = mkName "a"
     b = mkName "b"
 
-     -- GHC 7.6 uses containers-0.5.0.0 which doesn't have a good Data instance
-     -- for Map. So we have to convert to lists before comparing.
-    eq = (==) `on` fmap M.toList
-
 -- Test that type synonym expansion is efficient
 test_t123 :: ()
 test_t123 =
@@ -555,7 +629,10 @@ main = hspec $ do
 
     it "expands type synonyms in type variable binders" $ test_t97
 
-    it "collects GADT record selectors correctly" $ test_t102
+    it "reifies GADT record selectors correctly" $ test_t100
+
+    zipWithM (\b n -> it ("collects GADT record selectors correctly" ++ show n) b)
+      test_t102 [1..]
 
     it "quantifies kind variables in desugared ADT constructors" $ test_t103
 
@@ -564,6 +641,8 @@ main = hspec $ do
     zipWithM (\b n -> it ("toposorts free variables deterministically " ++ show n) b)
       test_t112 [1..]
 
+    it "reifies fixity declarations inside of classes" $ test_t132
+
     zipWithM (\b n -> it ("computes free variables correctly " ++ show n) b)
       test_fvs [1..]
 
@@ -590,4 +669,13 @@ main = hspec $ do
     zipWithM (\b n -> it ("matches types " ++ show n) b)
       test_matchTy [1..]
 
+    zipWithM (\b n -> it ("reifies kinds of declarations with CUSKs " ++ show n) b)
+      test_reify_type_cusks [1..]
+
+    zipWithM (\b n -> it ("reifies kinds of declarations without CUSKs " ++ show n) b)
+      test_reify_type_no_cusks [1..]
+
+    zipWithM (\b n -> it ("reifies the kinds of declarations with signatures " ++ show n) b)
+      test_reify_kind_sigs [1..]
+
     fromHUnitTest tests
